#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""FIDUCEO MVIRI FCDR Reader.

Introduction
------------
The FIDUCEO MVIRI FCDR is a Fundamental Climate Data Record (FCDR) of
re-calibrated Level 1.5 Infrared, Water Vapour, and Visible radiances from
the Meteosat Visible Infra-Red Imager (MVIRI) instrument onboard the
Meteosat First Generation satellites. There are two variants of the dataset:
The *full FCDR* and a simplified version called *easy FCDR*. Some datasets are
only available in one of the two variants, see the corresponding YAML
definition in ``satpy/etc/readers/``.

Dataset Names
-------------
The FIDUCEO MVIRI readers use names ``VIS``, ``WV`` and ``IR`` for the visible,
water vapor and infrared channels, respectively. These are different from
the original netCDF variable names for the following reasons:

- VIS channel is named differently in full FCDR (``counts_vis``) and easy FCDR
  (``toa_bidirectional_reflectance_vis``)
- netCDF variable names contain the calibration level (e.g. ``counts_...``),
  which might be confusing for satpy users if a different calibration level
  is chosen.

Remaining datasets (such as quality flags and uncertainties) have the same
name in the reader as in the netCDF file.


Example:
--------
This is how to read FIDUCEO MVIRI FCDR data in satpy:

.. code-block:: python

    from satpy import Scene

    scn = Scene(filenames=['FIDUCEO_FCDR_L15_MVIRI_MET7-57.0...'],
                reader='mviri_l1b_fiduceo_nc')
    scn.load(['VIS', 'WV', 'IR'])

Global netCDF attributes are available in the ``raw_metadata`` attribute of
each loaded dataset.


Image Orientation
-----------------
The images are stored in MVIRI scanning direction, that means South is up and
East is right. This can be changed as follows:

.. code-block:: python

    scn.load(['VIS'], upper_right_corner='NE')


Geolocation
-----------
In addition to the image data, FIDUCEO also provides so called *static FCDRs*
containing latitude and longitude coordinates. In order to simplify their
usage, the FIDUCEO MVIRI readers do not make use of these static files, but
instead provide an area definition that can be used to compute longitude and
latitude coordinates on demand.

.. code-block:: python

    area = scn['VIS'].attrs['area']
    lons, lats = area.get_lonlats()

Those were compared to the static FCDR and they agree very well, however there
are small differences. The mean difference is < 1E3 degrees for all channels
and projection longitudes.


Huge VIS Reflectances
---------------------
You might encounter huge VIS reflectances (10^8 percent and greater) in
situations where both radiance and solar zenith angle are small. The reader
certainly needs some improvement in this regard. Maybe the corresponding
uncertainties can be used to filter these cases before calculating reflectances.


VIS Channel Quality Flags
-------------------------
Quality flags are available for the VIS channel only. A simple approach for
masking bad quality pixels is to set the ``mask_bad_quality`` keyword argument
to ``True``:

.. code-block:: python

    scn = Scene(filenames=['FIDUCEO_FCDR_L15_MVIRI_MET7-57.0...'],
                reader='mviri_l1b_fiduceo_nc',
                reader_kwargs={'mask_bad_quality': True})

See :class:`FiduceoMviriBase` for an argument description. In some situations
however the entire image can be flagged (look out for warnings). In that case
check out the ``quality_pixel_bitmask`` and ``data_quality_bitmask`` datasets
to find out why.


Angles
------
The FIDUCEO MVIRI FCDR provides satellite and solar angles on a coarse tiepoint
grid. By default these datasets will be interpolated to the higher VIS
resolution. This can be changed as follows:

.. code-block:: python

    scn.load(['solar_zenith_angle'], resolution=4500)

If you need the angles in both resolutions, use data queries:

.. code-block:: python

    from satpy import DataQuery

    query_vis = DataQuery(
        name='solar_zenith_angle',
        resolution=2250
    )
    query_ir = DataQuery(
        name='solar_zenith_angle',
        resolution=4500
    )
    scn.load([query_vis, query_ir])

    # Use the query objects to access the datasets as follows
    sza_vis = scn[query_vis]


References:
-----------
    - `[Handbook]`_ MFG User Handbook
    - `[PUG]`_ FIDUCEO MVIRI FCDR Product User Guide

.. _[Handbook]: https://www.eumetsat.int/media/7323
.. _[PUG]: http://doi.org/10.15770/EUM_SEC_CLM_0009
"""

import abc
import contextlib
import functools
import warnings

import dask.array as da
import numpy as np
import xarray as xr

from satpy.readers.core._geos_area import get_area_definition, get_area_extent, sampling_to_lfac_cfac
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_chunk_size_limit

CHUNK_SIZE = get_chunk_size_limit()
EQUATOR_RADIUS = 6378140.0
POLE_RADIUS = 6356755.0
ALTITUDE = 42164000.0 - EQUATOR_RADIUS
"""[Handbook] section 5.2.1."""

MVIRI_FIELD_OF_VIEW = 18.0
"""[Handbook] section 5.3.2.1."""

CHANNELS = ["VIS", "WV", "IR"]
ANGLES = [
    "solar_zenith_angle",
    "solar_azimuth_angle",
    "satellite_zenith_angle",
    "satellite_azimuth_angle"
]
OTHER_REFLECTANCES = [
    "u_independent_toa_bidirectional_reflectance",
    "u_structured_toa_bidirectional_reflectance"
]
HIGH_RESOL = 2250


class IRWVCalibrator:
    """Calibrate IR & WV channels."""

    def __init__(self, coefs):
        """Initialize the calibrator.

        Args:
            coefs: Calibration coefficients.
        """
        self.coefs = coefs

    def calibrate(self, counts, calibration):
        """Calibrate IR/WV counts to the given calibration."""
        if calibration == "counts":
            return counts
        elif calibration in ("radiance", "brightness_temperature"):
            return self._calibrate_rad_bt(counts, calibration)
        else:
            raise KeyError(
                "Invalid IR/WV calibration: {}".format(calibration.name)
            )

    def _calibrate_rad_bt(self, counts, calibration):
        """Calibrate counts to radiance or brightness temperature."""
        rad = self._counts_to_radiance(counts)
        if calibration == "radiance":
            return rad
        bt = self._radiance_to_brightness_temperature(rad)
        return bt

    def _counts_to_radiance(self, counts):
        """Convert IR/WV counts to radiance.

        Reference: [PUG], equations (4.1) and (4.2).
        """
        rad = self.coefs["a"] + self.coefs["b"] * counts
        return rad.where(rad > 0, np.float32(np.nan))

    def _radiance_to_brightness_temperature(self, rad):
        """Convert IR/WV radiance to brightness temperature.

        Reference: [PUG], equations (5.1) and (5.2).
        """
        bt = self.coefs["bt_b"] / (np.log(rad) - self.coefs["bt_a"])
        return bt.where(bt > 0, np.float32(np.nan))


class VISCalibrator:
    """Calibrate VIS channel."""

    def __init__(self, coefs, solar_zenith_angle=None):
        """Initialize the calibrator.

        Args:
            coefs:
                Calibration coefficients.
            solar_zenith_angle (numbers.Number, Optional):
                Solar zenith angle. Only required for calibration to
                reflectance.
        """
        self.coefs = coefs
        self.solar_zenith_angle = solar_zenith_angle

    def calibrate(self, counts, calibration):
        """Calibrate VIS counts."""
        if calibration == "counts":
            return counts
        elif calibration in ("radiance", "reflectance"):
            return self._calibrate_rad_refl(counts, calibration)
        else:
            raise KeyError(
                "Invalid VIS calibration: {}".format(calibration.name)
            )

    def _calibrate_rad_refl(self, counts, calibration):
        """Calibrate counts to radiance or reflectance."""
        rad = self._counts_to_radiance(counts)
        if calibration == "radiance":
            return rad
        refl = self._radiance_to_reflectance(rad)
        refl = self.update_refl_attrs(refl)
        return refl

    def _counts_to_radiance(self, counts):
        """Convert VIS counts to radiance.

        Reference: [PUG], equations (7) and (8).
        """
        years_since_launch = self.coefs["years_since_launch"]
        a_cf = (self.coefs["a0"] +
                self.coefs["a1"] * years_since_launch +
                self.coefs["a2"] * years_since_launch ** 2)
        mean_count_space_vis = self.coefs["mean_count_space"]
        rad = (counts - mean_count_space_vis) * a_cf
        return rad.where(rad > 0, np.float32(np.nan))

    def _radiance_to_reflectance(self, rad):
        """Convert VIS radiance to reflectance factor.

        Note: Produces huge reflectances in situations where both radiance and
        solar zenith angle are small. Maybe the corresponding uncertainties
        can be used to filter these cases before calculating reflectances.

        Reference: [PUG], equation (6).
        """
        sza = self.solar_zenith_angle.where(
            da.fabs(self.solar_zenith_angle) < 90,
            np.float32(np.nan)
        )  # direct illumination only
        cos_sza = np.cos(np.deg2rad(sza))
        refl = (
           (np.pi * self.coefs["distance_sun_earth"] ** 2) /
           (self.coefs["solar_irradiance"] * cos_sza) *
           rad
        )
        return self.refl_factor_to_percent(refl)

    def update_refl_attrs(self, refl):
        """Update attributes of reflectance datasets."""
        refl.attrs["sun_earth_distance_correction_applied"] = True
        refl.attrs["sun_earth_distance_correction_factor"] = self.coefs[
            "distance_sun_earth"].item()
        return refl

    @staticmethod
    def refl_factor_to_percent(refl):
        """Convert reflectance factor to percent."""
        return refl * 100


class Navigator:
    """Navigate MVIRI images."""

    def get_area_def(self, im_size, projection_longitude):
        """Create MVIRI area definition."""
        proj_params = self._get_proj_params(im_size, projection_longitude)
        extent = get_area_extent(proj_params)
        return get_area_definition(proj_params, extent)

    def _get_proj_params(self, im_size, projection_longitude):
        """Get projection parameters for the given settings."""
        area_name = "geos_mviri_{0}x{0}".format(im_size)
        lfac, cfac, loff, coff = self._get_factors_offsets(im_size)
        return {
            "ssp_lon": projection_longitude,
            "a": EQUATOR_RADIUS,
            "b": POLE_RADIUS,
            "h": ALTITUDE,
            "units": "m",
            "loff": loff - im_size,
            "coff": coff,
            "lfac": -lfac,
            "cfac": -cfac,
            "nlines": im_size,
            "ncols": im_size,
            "scandir": "S2N",  # Reference: [PUG] section 2.
            "p_id": area_name,
            "a_name": area_name,
            "a_desc": "MVIRI Geostationary Projection"
        }

    def _get_factors_offsets(self, im_size):
        """Determine line/column offsets and scaling factors."""
        # For offsets see variables "asamp" and "aline" of subroutine
        # "refgeo" in [Handbook] and in
        # https://github.com/FIDUCEO/FCDR_MVIRI/blob/master/lib/nrCrunch/cruncher.f
        loff = coff = im_size / 2 + 0.5
        lfac = cfac = sampling_to_lfac_cfac(
            np.deg2rad(MVIRI_FIELD_OF_VIEW) / im_size
        )
        return lfac, cfac, loff, coff


class Interpolator:
    """Interpolate datasets to another resolution."""

    @staticmethod
    def interp_tiepoints(ds, target_x, target_y):
        """Interpolate dataset between tiepoints.

        Uses linear interpolation.

        FUTURE: [PUG] recommends cubic spline interpolation.

        Args:
            ds:
                Dataset to be interpolated
            target_x:
                Target x coordinates
            target_y:
                Target y coordinates
        """
        # No tiepoint coordinates specified in the files. Use dimensions
        # to calculate tiepoint sampling and assign tiepoint coordinates
        # accordingly.
        sampling = target_x.size // ds.coords["x"].size
        ds = ds.assign_coords(x=target_x.values[::sampling],
                              y=target_y.values[::sampling])

        return ds.interp(x=target_x.values, y=target_y.values)

    @staticmethod
    def interp_acq_time(time2d, target_y):
        """Interpolate scanline acquisition time to the given coordinates.

        The files provide timestamps per pixel for the low resolution
        channels (IR/WV) only.

        1) Average values in each line to obtain one timestamp per line.
        2) For the VIS channel duplicate values in y-direction (as
           advised by [PUG]).

        Note that the timestamps do not increase monotonically with the
        line number in some cases.

        Returns:
            Mean scanline acquisition timestamps
        """
        # Compute mean timestamp per scanline
        time = time2d.mean(dim="x")

        # If required, repeat timestamps in y-direction to obtain higher
        # resolution
        y = time.coords["y"].values
        if y.size < target_y.size:
            reps = target_y.size // y.size
            y_rep = np.repeat(y, reps)
            time_hires = time.reindex(y=y_rep)
            time_hires = time_hires.assign_coords(y=target_y)
            return time_hires
        return time


class VisQualityControl:
    """Simple quality control for VIS channel."""

    def __init__(self, mask):
        """Initialize the quality control."""
        self._mask = mask

    def check(self):
        """Check VIS channel quality and issue a warning if it's bad."""
        use_with_caution = da.bitwise_and(self._mask, 2)
        if use_with_caution.all():
            warnings.warn(
                'All pixels of the VIS channel are flagged as "use with '
                'caution". Use datasets "quality_pixel_bitmask" and '
                '"data_quality_bitmask" to find out why.',
                stacklevel=2
            )

    def mask(self, ds):
        """Mask VIS pixels with bad quality.

        Pixels are considered bad quality if the "quality_pixel_bitmask" is
        everything else than 0 (no flag set).
        """
        return ds.where(self._mask == 0, np.float32(np.nan))


def is_high_resol(resolution):
    """Identify high resolution channel."""
    return resolution == HIGH_RESOL


def preprocess_dataset(ds):
    """Preprocess the given dataset.

    Performs steps that can be done once, such as decoding
    according to CF conventions.
    """
    preproc = _DatasetPreprocessor()
    return preproc.preprocess(ds)


class _DatasetPreprocessor:
    """Helper class for preprocessing the dataset."""

    def preprocess(self, ds):
        """Preprocess the given dataset."""
        ds = self._rename_vars(ds)
        ds = self._decode_cf(ds)
        ds = self._fix_duplicate_dimensions(ds)
        self._reassign_coords(ds)
        self._cleanup_attrs(ds)
        return ds

    def _rename_vars(self, ds):
        """Rename variables to match satpy's expectations."""
        new_names = {
            "time_ir_wv": "time",
        }
        new_names_avail = {
            old: new
            for old, new in new_names.items()
            if old in ds
        }
        return ds.rename(new_names_avail)

    def _decode_cf(self, ds):
        """Decode data according to CF conventions."""
        # CF decoding fails because time coordinate contains fill values.
        # Decode time separately, then decode rest using decode_cf().
        time = self._decode_time(ds)
        ds = ds.drop_vars(time.name)
        ds = xr.decode_cf(ds)
        ds[time.name] = (time.dims, time.values)
        return ds

    def _decode_time(self, ds):
        """Decode time using fill value and offset.

        Replace fill values with NaT.
        """
        time = ds["time"]
        time_dec = (time + time.attrs["add_offset"]).astype("datetime64[s]").astype("datetime64[ns]")
        is_fill_value = time == time.attrs["_FillValue"]
        return xr.where(is_fill_value, np.datetime64("NaT"), time_dec)

    def _fix_duplicate_dimensions(self, ds):
        """Rename dimensions as duplicate dimensions names are not supported by xarray."""
        ds = ds.copy()
        ds.variables["covariance_spectral_response_function_vis"].dims = ("srf_size_1", "srf_size_2")
        ds.variables["channel_correlation_matrix_independent"].dims = ("channel_1", "channel_2")
        ds.variables["channel_correlation_matrix_structured"].dims = ("channel_1", "channel_2")
        return ds.drop_dims(["channel", "srf_size"])

    def _reassign_coords(self, ds):
        """Re-assign coordinates.

        For some reason xarray doesn't assign coordinates to all high
        resolution data variables. In that case ds["varname"] doesn't
        have coords, but they're still in ds.coords.
        """
        for var_name, data_array in ds.data_vars.items():
            if self._coordinates_not_assigned(data_array):
                ds[var_name] = data_array.assign_coords(
                    {
                        "y": ds.coords["y"],
                        "x": ds.coords["x"]
                    }
                )

    def _coordinates_not_assigned(self, data_array):
        return "y" in data_array.dims and "y" not in data_array.coords

    def _cleanup_attrs(self, ds):
        """Cleanup dataset attributes."""
        # Remove ancillary_variables attribute to avoid downstream
        # satpy warnings.
        for data_array in ds.data_vars.values():
            data_array.attrs.pop("ancillary_variables", None)


class DatasetAccessor:
    """Helper class for accessing the dataset.

    Performs steps that need to be done each time a variable
    is accessed, such as renaming "y_*" coordinates to "y".
    """

    def __init__(self, ds):
        """Wrap the given dataset."""
        self.ds = ds

    @property
    def attrs(self):
        """Exposes dataset attributes."""
        return self.ds.attrs

    def __getitem__(self, item):
        """Get a variable from the dataset."""
        data_array = self.ds[item]
        if self._should_dims_be_renamed(data_array):
            return self._rename_dims(data_array)
        return data_array

    def _should_dims_be_renamed(self, data_array):
        """Determine whether dataset dimensions need to be renamed."""
        return "y_ir_wv" in data_array.dims or "y_tie" in data_array.dims

    def _rename_dims(self, data_array):
        """Rename dataset dimensions to match satpy's expectations."""
        new_names = {
            "y_ir_wv": "y",
            "x_ir_wv": "x",
            "y_tie": "y",
            "x_tie": "x"
        }
        new_names_avail = {
            old: new
            for old, new in new_names.items()
            if old in data_array.dims
        }
        return data_array.rename(new_names_avail)

    def get_xy_coords(self, resolution):
        """Get x and y coordinates for the given resolution."""
        if is_high_resol(resolution):
            return self.ds.coords["x"], self.ds.coords["y"]
        return self.ds.coords["x_ir_wv"], self.ds.coords["x_ir_wv"]

    def get_image_size(self, resolution):
        """Get image size for the given resolution."""
        if is_high_resol(resolution):
            return self.ds.coords["y"].size
        return self.ds.coords["y_ir_wv"].size


def open_dataset(filename):
    """Load dataset from the given file."""
    with ignore_dup_dim_warning():
        nc_raw = xr.open_dataset(
            filename,
            chunks={"x": CHUNK_SIZE,
                    "y": CHUNK_SIZE,
                    "x_ir_wv": CHUNK_SIZE,
                    "y_ir_wv": CHUNK_SIZE},
            # see dataset preprocessor for why decoding is disabled
            decode_cf=False,
            decode_times=False,
            mask_and_scale=False,
        )
        nc_preproc = preprocess_dataset(nc_raw)
    return DatasetAccessor(nc_preproc)


class FiduceoMviriBase(BaseFileHandler):
    """Baseclass for FIDUCEO MVIRI file handlers."""

    nc_keys = {
        "WV": "count_wv",
        "IR": "count_ir"
    }

    def __init__(self, filename, filename_info, filetype_info,  # noqa: D417
                 mask_bad_quality=False):
        """Initialize the file handler.

        Args:
             mask_bad_quality: Mask VIS pixels with bad quality, that means
                 any quality flag except "ok". If you need more control, use
                 the ``quality_pixel_bitmask`` and ``data_quality_bitmask``
                 datasets.
        """
        super(FiduceoMviriBase, self).__init__(
            filename, filename_info, filetype_info)
        self.mask_bad_quality = mask_bad_quality
        self.nc = open_dataset(filename)
        self.projection_longitude = self._get_projection_longitude(filename_info)
        self.calib_coefs = self._get_calib_coefs()
        self._get_angles = functools.lru_cache(maxsize=8)(
            self._get_angles_uncached
        )
        self._get_acq_time = functools.lru_cache(maxsize=3)(
            self._get_acq_time_uncached
        )

    def _get_projection_longitude(self, filename_info):
        """Read projection longitude from filename as it is not provided in the file."""
        if "." in str(filename_info["projection_longitude"]):
            return float(filename_info["projection_longitude"])

        return float(filename_info["projection_longitude"]) / 100

    def get_dataset(self, dataset_id, dataset_info):
        """Get the dataset."""
        name = dataset_id["name"]
        resolution = dataset_id["resolution"]
        if name in ANGLES:
            ds = self._get_angles(name, resolution)
        elif name in CHANNELS:
            ds = self._get_channel(name, resolution, dataset_id["calibration"])
        else:
            ds = self._get_other_dataset(name)
        ds = self._cleanup_coords(ds)
        self._update_attrs(ds, dataset_info)
        return ds

    def get_area_def(self, dataset_id):
        """Get area definition of the given dataset."""
        im_size = self.nc.get_image_size(dataset_id["resolution"])
        nav = Navigator()
        return nav.get_area_def(
            im_size=im_size,
            projection_longitude=self.projection_longitude
        )

    def _get_channel(self, name, resolution, calibration):
        """Get and calibrate channel data."""
        ds = self.nc[self.nc_keys[name]]
        ds = self._calibrate(
            ds,
            channel=name,
            calibration=calibration
        )
        if name == "VIS":
            qc = VisQualityControl(self.nc["quality_pixel_bitmask"])
            if self.mask_bad_quality:
                ds = qc.mask(ds)
            else:
                qc.check()
        ds["acq_time"] = self._get_acq_time(resolution)
        return ds

    def _get_angles_uncached(self, name, resolution):
        """Get angle dataset.

        Files provide angles (solar/satellite zenith & azimuth) at a coarser
        resolution. Interpolate them to the desired resolution.
        """
        angles = self.nc[name]
        target_x, target_y = self.nc.get_xy_coords(resolution)
        return Interpolator.interp_tiepoints(
            angles,
            target_x=target_x,
            target_y=target_y
        )

    def _get_other_dataset(self, name):
        """Get other datasets such as uncertainties."""
        ds = self.nc[name]
        if name in OTHER_REFLECTANCES:
            ds = VISCalibrator.refl_factor_to_percent(ds)
        return ds

    def _update_attrs(self, ds, info):
        """Update dataset attributes."""
        ds.attrs.update(info)
        ds.attrs.update({"platform": self.filename_info["platform"],
                         "sensor": self.filename_info["sensor"]})
        ds.attrs["raw_metadata"] = self.nc.attrs
        ds.attrs["orbital_parameters"] = self._get_orbital_parameters()

    def _cleanup_coords(self, ds):
        """Cleanup dataset coordinates.

        Y/x coordinates have been useful for interpolation so far, but they
        only contain row/column numbers. Drop these coordinates so that Satpy
        can assign projection coordinates upstream (based on the area
        definition).
        """
        return ds.drop_vars(["y", "x"])

    def _calibrate(self, ds, channel, calibration):
        """Calibrate the given dataset."""
        if channel == "VIS":
            return self._calibrate_vis(ds, channel, calibration)
        calib = IRWVCalibrator(self.calib_coefs[channel])
        return calib.calibrate(ds, calibration)

    @abc.abstractmethod
    def _calibrate_vis(self, ds, channel, calibration):  # pragma: no cover
        """Calibrate VIS channel. To be implemented by subclasses."""
        raise NotImplementedError

    def _get_calib_coefs(self):
        """Get calibration coefficients for all channels.

        Note: Only coefficients present in both file types.
        """
        coefs = {
            "VIS": {
                "distance_sun_earth": self.nc["distance_sun_earth"],
                "solar_irradiance": self.nc["solar_irradiance_vis"]
            },
            "IR": {
                "a": self.nc["a_ir"],
                "b": self.nc["b_ir"],
                "bt_a": self.nc["bt_a_ir"],
                "bt_b": self.nc["bt_b_ir"]
            },
            "WV": {
                "a": self.nc["a_wv"],
                "b": self.nc["b_wv"],
                "bt_a": self.nc["bt_a_wv"],
                "bt_b": self.nc["bt_b_wv"]
            },
        }

        # Convert coefficients to 32bit float to reduce memory footprint
        # of calibrated data.
        for ch in coefs:
            for name in coefs[ch]:
                coefs[ch][name] = np.float32(coefs[ch][name])

        return coefs

    def _get_acq_time_uncached(self, resolution):
        """Get scanline acquisition time for the given resolution.

        Note that the acquisition time does not increase monotonically
        with the scanline number due to the scan pattern and rectification.
        """
        time2d = self.nc["time"]
        _, target_y = self.nc.get_xy_coords(resolution)
        return Interpolator.interp_acq_time(time2d, target_y=target_y.values)

    def _get_orbital_parameters(self):
        """Get the orbital parameters."""
        orbital_parameters = {
            "projection_longitude": self.projection_longitude,
            "projection_latitude": 0.0,
            "projection_altitude": ALTITUDE
        }
        ssp_lon, ssp_lat = self._get_ssp_lonlat()
        if not np.isnan(ssp_lon) and not np.isnan(ssp_lat):
            orbital_parameters.update({
                "satellite_actual_longitude": ssp_lon,
                "satellite_actual_latitude": ssp_lat,
                # altitude not available
            })
        return orbital_parameters

    def _get_ssp_lonlat(self):
        """Get longitude and latitude at the subsatellite point.

        Easy FCDR files provide satellite position at the beginning and
        end of the scan. This method computes the mean of those two values.
        In the full FCDR the information seems to be missing.

        Returns:
            Subsatellite longitude and latitude
        """
        ssp_lon = self._get_ssp("longitude")
        ssp_lat = self._get_ssp("latitude")
        return ssp_lon, ssp_lat

    def _get_ssp(self, coord):
        key_start = "sub_satellite_{}_start".format(coord)
        key_end = "sub_satellite_{}_end".format(coord)
        try:
            sub_lonlat = np.nanmean(
                [self.nc[key_start].values,
                 self.nc[key_end].values]
            )
        except KeyError:
            # Variables seem to be missing in full FCDR
            sub_lonlat = np.nan
        return sub_lonlat


class FiduceoMviriEasyFcdrFileHandler(FiduceoMviriBase):
    """File handler for FIDUCEO MVIRI Easy FCDR."""

    nc_keys = FiduceoMviriBase.nc_keys.copy()
    nc_keys["VIS"] = "toa_bidirectional_reflectance_vis"

    def _calibrate_vis(self, ds, channel, calibration):
        """Calibrate VIS channel.

        Easy FCDR provides reflectance only, no counts or radiance.
        """
        if calibration == "reflectance":
            coefs = self.calib_coefs[channel]
            cal = VISCalibrator(coefs)
            refl = cal.refl_factor_to_percent(ds)
            refl = cal.update_refl_attrs(refl)
            return refl
        elif calibration in ("counts", "radiance"):
            raise ValueError("Cannot calibrate to {}. Easy FCDR provides "
                             "reflectance only.".format(calibration.name))
        else:
            raise KeyError("Invalid calibration: {}".format(calibration.name))


class FiduceoMviriFullFcdrFileHandler(FiduceoMviriBase):
    """File handler for FIDUCEO MVIRI Full FCDR."""

    nc_keys = FiduceoMviriBase.nc_keys.copy()
    nc_keys["VIS"] = "count_vis"

    def _get_calib_coefs(self):
        """Add additional VIS coefficients only present in full FCDR."""
        coefs = super()._get_calib_coefs()
        coefs["VIS"].update({
            "years_since_launch": np.float32(self.nc["years_since_launch"]),
            "a0": np.float32(self.nc["a0_vis"]),
            "a1": np.float32(self.nc["a1_vis"]),
            "a2": np.float32(self.nc["a2_vis"]),
            "mean_count_space": np.float32(
                self.nc["mean_count_space_vis"]
            )
        })
        return coefs

    def _calibrate_vis(self, ds, channel, calibration):
        """Calibrate VIS channel."""
        sza = None
        if calibration == "reflectance":
            sza = self._get_angles("solar_zenith_angle", HIGH_RESOL)
        cal = VISCalibrator(self.calib_coefs[channel], sza)
        return cal.calibrate(ds, calibration)


@contextlib.contextmanager
def ignore_dup_dim_warning():
    """Ignore the expected warning from xarray when working with duplicate dimensions."""
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore",
                                message=".*We do not yet support duplicate dimension names.*")
        yield



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Test for ancillary data downloading."""

from unittest import mock

import pytest
import yaml

from satpy.aux_download import DataDownloadMixin
from satpy.modifiers import ModifierBase

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmpdir

pooch = pytest.importorskip("pooch")

README_URL = "https://raw.githubusercontent.com/pytroll/satpy/main/README.rst"


class UnfriendlyModifier(ModifierBase, DataDownloadMixin):
    """Fake modifier that raises an exception in __init__."""

    def __init__(self, name, prerequisites=None, optional_prerequisites=None, **kwargs):
        """Raise an exception if we weren't provided any prerequisites."""
        if not prerequisites or len(prerequisites) != 1:
            raise ValueError("Unexpected number of prereqs")
        super().__init__(name, prerequisites, optional_prerequisites, **kwargs)
        self.register_data_files({"url": kwargs["url"],
                                  "filename": kwargs["filename"],
                                  "known_hash": kwargs["known_hash"]})


def _setup_custom_composite_config(base_dir):
    from satpy.composites.aux_data import StaticImageCompositor
    from satpy.modifiers.atmosphere import ReflectanceCorrector
    composite_config = base_dir.mkdir("composites").join("visir.yaml")
    with open(composite_config, "w") as comp_file:
        yaml.dump({
            "sensor_name": "visir",
            "modifiers": {
                "test_modifier": {
                    "modifier": ReflectanceCorrector,
                    "url": README_URL,
                    "known_hash": None,
                },
                "unfriendly_modifier": {
                    "modifier": UnfriendlyModifier,
                    "url": README_URL,
                    "filename": "unfriendly.rst",
                    "known_hash": None,
                }
            },
            "composites": {
                "test_static": {
                    "compositor": StaticImageCompositor,
                    "url": README_URL,
                    "known_hash": None,
                },
            },
        }, comp_file)


def _setup_custom_reader_config(base_dir):
    reader_config = base_dir.mkdir("readers").join("fake.yaml")
    with open(reader_config, "wt") as comp_file:
        # abstract base classes can't be converted so we do raw string
        comp_file.write("""
reader:
  name: "fake"
  reader: !!python/name:satpy.readers.core.yaml_reader.FileYAMLReader
  data_files:
    - url: {}
      known_hash: null
    - url: {}
      filename: "README2.rst"
      known_hash: null
file_types: {{}}
""".format(README_URL, README_URL))


def _setup_custom_writer_config(base_dir):
    writer_config = base_dir.mkdir("writers").join("fake.yaml")
    with open(writer_config, "wt") as comp_file:
        # abstract base classes can't be converted so we do raw string
        comp_file.write("""
writer:
  name: "fake"
  writer: !!python/name:satpy.writers.core.base.Writer
  data_files:
    - url: {}
      known_hash: null
    - url: {}
      filename: "README2.rst"
      known_hash: null
""".format(README_URL, README_URL))


def _assert_reader_files_downloaded(readers, found_files):
    r_cond1 = "readers/README.rst" in found_files
    r_cond2 = "readers/README2.rst" in found_files
    if readers is not None and not readers:
        r_cond1 = not r_cond1
        r_cond2 = not r_cond2
    assert r_cond1
    assert r_cond2


def _assert_writer_files_downloaded(writers, found_files):
    w_cond1 = "writers/README.rst" in found_files
    w_cond2 = "writers/README2.rst" in found_files
    if writers is not None and not writers:
        w_cond1 = not w_cond1
        w_cond2 = not w_cond2
    assert w_cond1
    assert w_cond2


def _assert_comp_files_downloaded(comp_sensors, found_files):
    comp_cond = "composites/README.rst" in found_files
    if comp_sensors is not None and not comp_sensors:
        comp_cond = not comp_cond
    assert comp_cond


def _assert_mod_files_downloaded(comp_sensors, found_files):
    mod_cond = "modifiers/README.rst" in found_files
    unfriendly_cond = "modifiers/unfriendly.rst" in found_files
    if comp_sensors is not None and not comp_sensors:
        mod_cond = not mod_cond
    assert mod_cond
    assert not unfriendly_cond


class TestDataDownload:
    """Test basic data downloading functionality."""

    @pytest.fixture(autouse=True)
    def _setup_custom_configs(self, tmpdir):
        _setup_custom_composite_config(tmpdir)
        _setup_custom_reader_config(tmpdir)
        _setup_custom_writer_config(tmpdir)
        self.tmpdir = tmpdir

    @pytest.mark.parametrize("comp_sensors", [tuple(), None, ("visir",)])
    @pytest.mark.parametrize("writers", [[], None, ["fake"]])
    @pytest.mark.parametrize("readers", [[], None, ["fake"]])
    def test_find_registerable(self, readers, writers, comp_sensors):
        """Test that find_registerable finds some things."""
        import satpy
        from satpy.aux_download import find_registerable_files
        with satpy.config.set(config_path=[self.tmpdir]), \
             mock.patch("satpy.aux_download._FILE_REGISTRY", {}):
            found_files = find_registerable_files(
                readers=readers, writers=writers,
                composite_sensors=comp_sensors,
            )

            _assert_reader_files_downloaded(readers, found_files)
            _assert_writer_files_downloaded(writers, found_files)
            _assert_comp_files_downloaded(comp_sensors, found_files)
            _assert_mod_files_downloaded(comp_sensors, found_files)

    def test_limited_find_registerable(self):
        """Test that find_registerable doesn't find anything when limited."""
        import satpy
        from satpy.aux_download import find_registerable_files
        file_registry = {}
        with satpy.config.set(config_path=[self.tmpdir]), \
             mock.patch("satpy.aux_download._FILE_REGISTRY", file_registry):
            found_files = find_registerable_files(
                readers=[], writers=[], composite_sensors=[],
            )
            assert not found_files

    def test_retrieve(self):
        """Test retrieving a single file."""
        import satpy
        from satpy.aux_download import find_registerable_files, retrieve
        file_registry = {}
        with satpy.config.set(config_path=[self.tmpdir], data_dir=str(self.tmpdir)), \
             mock.patch("satpy.aux_download._FILE_REGISTRY", file_registry):
            comp_file = "composites/README.rst"
            found_files = find_registerable_files()
            assert comp_file in found_files
            assert not self.tmpdir.join(comp_file).exists()
            retrieve(comp_file)
            assert self.tmpdir.join(comp_file).exists()

    def test_offline_retrieve(self):
        """Test retrieving a single file when offline."""
        import satpy
        from satpy.aux_download import find_registerable_files, retrieve
        file_registry = {}
        with satpy.config.set(config_path=[self.tmpdir], data_dir=str(self.tmpdir), download_aux=True), \
             mock.patch("satpy.aux_download._FILE_REGISTRY", file_registry):
            comp_file = "composites/README.rst"
            found_files = find_registerable_files()
            assert comp_file in found_files

            # the file doesn't exist, we can't download it
            assert not self.tmpdir.join(comp_file).exists()
            with satpy.config.set(download_aux=False):
                pytest.raises(RuntimeError, retrieve, comp_file)

            # allow downloading and get it
            retrieve(comp_file)
            assert self.tmpdir.join(comp_file).exists()

            # turn off downloading and make sure we get local file
            with satpy.config.set(download_aux=False):
                local_file = retrieve(comp_file)
                assert local_file

    def test_offline_retrieve_all(self):
        """Test registering and retrieving all files fails when offline."""
        import satpy
        from satpy.aux_download import retrieve_all
        with satpy.config.set(config_path=[self.tmpdir], data_dir=str(self.tmpdir), download_aux=False):
            pytest.raises(RuntimeError, retrieve_all)

    def test_retrieve_all(self):
        """Test registering and retrieving all files."""
        import satpy
        from satpy.aux_download import retrieve_all
        file_registry = {}
        file_urls = {}
        with satpy.config.set(config_path=[self.tmpdir], data_dir=str(self.tmpdir)), \
             mock.patch("satpy.aux_download._FILE_REGISTRY", file_registry), \
             mock.patch("satpy.aux_download._FILE_URLS", file_urls), \
             mock.patch("satpy.aux_download.find_registerable_files"):
            comp_file = "composites/README.rst"
            file_registry[comp_file] = None
            file_urls[comp_file] = README_URL
            assert not self.tmpdir.join(comp_file).exists()
            retrieve_all()
            assert self.tmpdir.join(comp_file).exists()

    def test_no_downloads_in_tests(self):
        """Test that tests aren't allowed to download stuff."""
        import satpy
        from satpy.aux_download import register_file, retrieve

        file_registry = {}
        with satpy.config.set(config_path=[self.tmpdir], data_dir=str(self.tmpdir),
                              download_aux=True), \
             mock.patch("satpy.aux_download._FILE_REGISTRY", file_registry):
            cache_key = "myfile.rst"
            register_file(README_URL, cache_key)
            assert not self.tmpdir.join(cache_key).exists()
            pytest.raises(RuntimeError, retrieve, cache_key)
            # touch the file so it gets created
            open(self.tmpdir.join(cache_key), "w").close()
            # offline downloading should still be allowed
            with satpy.config.set(download_aux=False):
                retrieve(cache_key)

    def test_download_script(self):
        """Test basic functionality of the download script."""
        import satpy
        from satpy.aux_download import retrieve_all_cmd
        file_registry = {}
        file_urls = {}
        with satpy.config.set(config_path=[self.tmpdir]), \
             mock.patch("satpy.aux_download._FILE_REGISTRY", file_registry), \
             mock.patch("satpy.aux_download._FILE_URLS", file_urls), \
             mock.patch("satpy.aux_download.find_registerable_files"):
            comp_file = "composites/README.rst"
            file_registry[comp_file] = None
            file_urls[comp_file] = README_URL
            assert not self.tmpdir.join(comp_file).exists()
            retrieve_all_cmd(argv=["--data-dir", str(self.tmpdir)])
            assert self.tmpdir.join(comp_file).exists()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for CF-compatible attributes encoding."""
import json


class TestCFAttributeEncoding:
    """Test case for CF attribute encodings."""

    def test__encode_nc_attrs(self):
        """Test attributes encoding."""
        from satpy.cf.attrs import encode_attrs_to_cf
        from satpy.tests.cf_tests._test_data import get_test_attrs
        from satpy.tests.utils import assert_dict_array_equality

        attrs, expected, _ = get_test_attrs()

        # Test encoding
        encoded = encode_attrs_to_cf(attrs)
        assert_dict_array_equality(expected, encoded)

        # Test decoding of json-encoded attributes
        raw_md_roundtrip = {"recarray": [[0, 0], [0, 0], [0, 0]],
                            "flag": "true",
                            "dict": {"a": 1, "b": [1, 2, 3]}}
        assert json.loads(encoded["raw_metadata"]) == raw_md_roundtrip
        assert json.loads(encoded["array_3d"]) == [[[1, 2], [3, 4]], [[1, 2], [3, 4]]]
        assert json.loads(encoded["nested_dict"]) == {"l1": {"l2": {"l3": [1, 2, 3]}}}
        assert json.loads(encoded["nested_list"]) == ["1", ["2", [3]]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 - 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
"""Tests for the CREFL ReflectanceCorrector modifier."""
import datetime as dt
from contextlib import contextmanager
from unittest import mock

import numpy as np
import pytest
import xarray as xr
from dask import array as da
from pyresample.geometry import AreaDefinition

from satpy.tests.utils import assert_maximum_dask_computes

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmpdir


@contextmanager
def mock_cmgdem(tmpdir, url):
    """Create fake file representing CMGDEM.hdf."""
    yield from _mock_and_create_dem_file(tmpdir, url, "averaged elevation", fill_value=-9999)


@contextmanager
def mock_tbase(tmpdir, url):
    """Create fake file representing tbase.hdf."""
    yield from _mock_and_create_dem_file(tmpdir, url, "Elevation")


def _mock_and_create_dem_file(tmpdir, url, var_name, fill_value=None):
    if not url:
        yield None
        return

    rmock_obj, dem_fn = _mock_dem_retrieve(tmpdir, url)
    _create_fake_dem_file(dem_fn, var_name, fill_value)

    try:
        yield rmock_obj
    finally:
        rmock_obj.stop()


def _mock_dem_retrieve(tmpdir, url):
    rmock_obj = mock.patch("satpy.modifiers._crefl.retrieve")
    rmock = rmock_obj.start()
    dem_fn = str(tmpdir.join(url))
    rmock.return_value = dem_fn
    return rmock_obj, dem_fn


def _create_fake_dem_file(dem_fn, var_name, fill_value):
    from pyhdf.SD import SD, SDC
    h = SD(dem_fn, SDC.WRITE | SDC.CREATE)
    dem_var = h.create(var_name, SDC.INT16, (10, 10))
    dem_var[:] = np.zeros((10, 10), dtype=np.int16)
    if fill_value is not None:
        dem_var.setfillvalue(fill_value)
    h.end()


def _make_viirs_xarray(data, area, name, standard_name, wavelength=None, units="degrees", calibration=None):
    return xr.DataArray(data, dims=("y", "x"),
                        attrs={
                            "start_orbit": 1708, "end_orbit": 1708, "wavelength": wavelength,
                            "modifiers": None, "calibration": calibration,
                            "resolution": 371, "name": name,
                            "standard_name": standard_name, "platform_name": "Suomi-NPP",
                            "polarization": None, "sensor": "viirs", "units": units,
                            "start_time": dt.datetime(2012, 2, 25, 18, 1, 24, 570942),
                            "end_time": dt.datetime(2012, 2, 25, 18, 11, 21, 175760), "area": area,
                            "ancillary_variables": []
                        })


class TestReflectanceCorrectorModifier:
    """Test the CREFL modifier."""

    @staticmethod
    def data_area_ref_corrector():
        """Create test area definition and data."""
        rows = 3
        cols = 5
        area = AreaDefinition(
            "some_area_name", "On-the-fly area", "geosabii",
            {"a": "6378137.0", "b": "6356752.31414", "h": "35786023.0", "lon_0": "-89.5", "proj": "geos", "sweep": "x",
             "units": "m"},
            cols, rows,
            (-5434894.954752679, -5434894.964451744, 5434894.964451744, 5434894.954752679))

        data = np.zeros((rows, cols)) + 25
        data[1, :] += 25
        data[2, :] += 50
        data = da.from_array(data, chunks=2)
        return area, data

    @pytest.mark.parametrize(
        ("name", "wavelength", "resolution", "exp_mean", "exp_unique"),
        [
            ("C01", (0.45, 0.47, 0.49), 1000, 44.757951,
             np.array([12.83774603, 14.38767557, 17.24258084, 41.87806142, 44.42472192, 47.89958451,
                       48.23343427, 48.53847386, 71.52916035, 72.26078684, 73.10523784])),
            ("C02", (0.59, 0.64, 0.69), 500, 51.4901,
             np.array([23.69999579, 24.00407203, 24.49390685, 51.4304448, 51.64271324, 51.70519738,
                       51.70942859, 51.76064747, 78.37182815, 78.77078522, 78.80199923])),
            ("C03", (0.8455, 0.865, 0.8845), 1000, 50.7243,
             np.array([24.78444631, 24.86790679, 24.99481254, 50.69670516, 50.72983327, 50.73601728,
                       50.75685498, 50.83136276, 76.39973287, 76.5714688, 76.59856607])),
            # ("C04", (1.3705, 1.378, 1.3855), 2000, 55.973458829136796, None),
            ("C05", (1.58, 1.61, 1.64), 1000, 52.7231,
             np.array([26.26568157, 26.43230852, 26.48936244, 52.00527783, 52.13043172, 52.20176747,
                       53.01505657, 53.29017112, 78.93907987, 79.49089239, 79.69387535])),
            ("C06", (2.225, 2.25, 2.275), 2000, 55.9735,
             np.array([27.82291562, 28.2268102, 28.37246323, 54.33639308, 54.61451818, 54.77543748,
                       56.62284858, 57.27288821, 83.57235975, 84.81324822, 85.27816457])),
        ]
    )
    def test_reflectance_corrector_abi(self, name, wavelength, resolution, exp_mean, exp_unique):
        """Test ReflectanceCorrector modifier with ABI data."""
        from satpy.modifiers._crefl import ReflectanceCorrector
        from satpy.tests.utils import make_dsq
        ref_cor = ReflectanceCorrector(optional_prerequisites=[
            make_dsq(name="satellite_azimuth_angle"),
            make_dsq(name="satellite_zenith_angle"),
            make_dsq(name="solar_azimuth_angle"),
            make_dsq(name="solar_zenith_angle")], name=name, prerequisites=[],
                                       wavelength=wavelength, resolution=resolution, calibration="reflectance",
                                       modifiers=("sunz_corrected", "rayleigh_corrected_crefl",), sensor="abi")

        assert ref_cor.attrs["modifiers"] == ("sunz_corrected", "rayleigh_corrected_crefl")
        assert ref_cor.attrs["calibration"] == "reflectance"
        assert ref_cor.attrs["wavelength"] == wavelength
        assert ref_cor.attrs["name"] == name
        assert ref_cor.attrs["resolution"] == resolution
        assert ref_cor.attrs["sensor"] == "abi"
        assert ref_cor.attrs["prerequisites"] == []
        assert ref_cor.attrs["optional_prerequisites"] == [
            make_dsq(name="satellite_azimuth_angle"),
            make_dsq(name="satellite_zenith_angle"),
            make_dsq(name="solar_azimuth_angle"),
            make_dsq(name="solar_zenith_angle")]

        area, dnb = self.data_area_ref_corrector()
        c01 = xr.DataArray(dnb,
                           dims=("y", "x"),
                           attrs={
                               "platform_name": "GOES-16",
                               "calibration": "reflectance", "units": "%", "wavelength": wavelength,
                               "name": name, "resolution": resolution, "sensor": "abi",
                               "start_time": "2017-09-20 17:30:40.800000", "end_time": "2017-09-20 17:41:17.500000",
                               "area": area, "ancillary_variables": [],
                               "orbital_parameters": {
                                   "satellite_nominal_longitude": -89.5,
                                   "satellite_nominal_latitude": 0.0,
                                   "satellite_nominal_altitude": 35786023.4375,
                               },
                           })
        with assert_maximum_dask_computes(0):
            res = ref_cor([c01], [])

        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["modifiers"] == ("sunz_corrected", "rayleigh_corrected_crefl")
        assert res.attrs["platform_name"] == "GOES-16"
        assert res.attrs["calibration"] == "reflectance"
        assert res.attrs["units"] == "%"
        assert res.attrs["wavelength"] == wavelength
        assert res.attrs["name"] == name
        assert res.attrs["resolution"] == resolution
        assert res.attrs["sensor"] == "abi"
        assert res.attrs["start_time"] == "2017-09-20 17:30:40.800000"
        assert res.attrs["end_time"] == "2017-09-20 17:41:17.500000"
        assert res.attrs["area"] == area
        assert res.attrs["ancillary_variables"] == []
        data = res.values
        unique = np.unique(data[~np.isnan(data)])
        np.testing.assert_allclose(np.nanmean(data), exp_mean, rtol=1e-5)
        assert data.shape == (3, 5)
        np.testing.assert_allclose(unique, exp_unique, rtol=1e-5)

    @pytest.mark.parametrize(
        ("url", "dem_mock_cm", "dem_sds"),
        [
            (None, mock_cmgdem, "average elevation"),
            ("CMGDEM.hdf", mock_cmgdem, "averaged elevation"),
            ("tbase.hdf", mock_tbase, "Elevation"),
        ])
    def test_reflectance_corrector_viirs(self, tmpdir, url, dem_mock_cm, dem_sds):
        """Test ReflectanceCorrector modifier with VIIRS data."""
        from satpy.modifiers._crefl import ReflectanceCorrector
        from satpy.tests.utils import make_dsq

        ref_cor = ReflectanceCorrector(
            optional_prerequisites=[
                make_dsq(name="satellite_azimuth_angle"),
                make_dsq(name="satellite_zenith_angle"),
                make_dsq(name="solar_azimuth_angle"),
                make_dsq(name="solar_zenith_angle")
            ],
            name="I01",
            prerequisites=[],
            wavelength=(0.6, 0.64, 0.68),
            resolution=371,
            calibration="reflectance",
            modifiers=("sunz_corrected_iband", "rayleigh_corrected_crefl_iband"),
            sensor="viirs",
            url=url,
            dem_sds=dem_sds,
        )

        assert ref_cor.attrs["modifiers"] == ("sunz_corrected_iband", "rayleigh_corrected_crefl_iband")
        assert ref_cor.attrs["calibration"] == "reflectance"
        assert ref_cor.attrs["wavelength"] == (0.6, 0.64, 0.68)
        assert ref_cor.attrs["name"] == "I01"
        assert ref_cor.attrs["resolution"] == 371
        assert ref_cor.attrs["sensor"] == "viirs"
        assert ref_cor.attrs["prerequisites"] == []
        assert ref_cor.attrs["optional_prerequisites"] == [
            make_dsq(name="satellite_azimuth_angle"),
            make_dsq(name="satellite_zenith_angle"),
            make_dsq(name="solar_azimuth_angle"),
            make_dsq(name="solar_zenith_angle")]

        area, data = self.data_area_ref_corrector()
        c01 = _make_viirs_xarray(data, area, "I01", "toa_bidirectional_reflectance",
                                 wavelength=(0.6, 0.64, 0.68), units="%",
                                 calibration="reflectance")
        c02 = _make_viirs_xarray(data, area, "satellite_azimuth_angle", "sensor_azimuth_angle")
        c03 = _make_viirs_xarray(data, area, "satellite_zenith_angle", "sensor_zenith_angle")
        c04 = _make_viirs_xarray(data, area, "solar_azimuth_angle", "solar_azimuth_angle")
        c05 = _make_viirs_xarray(data, area, "solar_zenith_angle", "solar_zenith_angle")

        with dem_mock_cm(tmpdir, url), assert_maximum_dask_computes(0):
            res = ref_cor([c01], [c02, c03, c04, c05])

        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["wavelength"] == (0.6, 0.64, 0.68)
        assert res.attrs["modifiers"] == ("sunz_corrected_iband", "rayleigh_corrected_crefl_iband")
        assert res.attrs["calibration"] == "reflectance"
        assert res.attrs["resolution"] == 371
        assert res.attrs["name"] == "I01"
        assert res.attrs["standard_name"] == "toa_bidirectional_reflectance"
        assert res.attrs["platform_name"] == "Suomi-NPP"
        assert res.attrs["sensor"] == "viirs"
        assert res.attrs["units"] == "%"
        assert res.attrs["start_time"] == dt.datetime(2012, 2, 25, 18, 1, 24, 570942)
        assert res.attrs["end_time"] == dt.datetime(2012, 2, 25, 18, 11, 21, 175760)
        assert res.attrs["area"] == area
        assert res.attrs["ancillary_variables"] == []
        data = res.values
        assert abs(np.mean(data) - 51.12750267805715) < 1e-6
        assert data.shape == (3, 5)
        unique = np.unique(data)
        np.testing.assert_allclose(unique, [25.20341703, 52.38819447, 75.79089654])

    def test_reflectance_corrector_modis(self):
        """Test ReflectanceCorrector modifier with MODIS data."""
        from satpy.modifiers._crefl import ReflectanceCorrector
        from satpy.tests.utils import make_dsq
        sataa_did = make_dsq(name="satellite_azimuth_angle")
        satza_did = make_dsq(name="satellite_zenith_angle")
        solaa_did = make_dsq(name="solar_azimuth_angle")
        solza_did = make_dsq(name="solar_zenith_angle")
        ref_cor = ReflectanceCorrector(
            optional_prerequisites=[sataa_did, satza_did, solaa_did, solza_did], name="1",
            prerequisites=[], wavelength=(0.62, 0.645, 0.67), resolution=250, calibration="reflectance",
            modifiers=("sunz_corrected", "rayleigh_corrected_crefl"), sensor="modis")
        assert ref_cor.attrs["modifiers"] == ("sunz_corrected", "rayleigh_corrected_crefl")
        assert ref_cor.attrs["calibration"] == "reflectance"
        assert ref_cor.attrs["wavelength"] == (0.62, 0.645, 0.67)
        assert ref_cor.attrs["name"] == "1"
        assert ref_cor.attrs["resolution"] == 250
        assert ref_cor.attrs["sensor"] == "modis"
        assert ref_cor.attrs["prerequisites"] == []
        assert ref_cor.attrs["optional_prerequisites"] == [
            make_dsq(name="satellite_azimuth_angle"),
            make_dsq(name="satellite_zenith_angle"),
            make_dsq(name="solar_azimuth_angle"),
            make_dsq(name="solar_zenith_angle")]

        area, dnb = self.data_area_ref_corrector()

        def make_xarray(name, calibration, wavelength=None, modifiers=None, resolution=1000):
            return xr.DataArray(dnb,
                                dims=("y", "x"),
                                attrs={
                                    "wavelength": wavelength, "level": None, "modifiers": modifiers,
                                    "calibration": calibration, "resolution": resolution,
                                    "name": name, "coordinates": ["longitude", "latitude"],
                                    "platform_name": "EOS-Aqua", "polarization": None, "sensor": "modis",
                                    "units": "%", "start_time": dt.datetime(2012, 8, 13, 18, 46, 1, 439838),
                                    "end_time": dt.datetime(2012, 8, 13, 18, 57, 47, 746296), "area": area,
                                    "ancillary_variables": []
                                })

        c01 = make_xarray("1", "reflectance", wavelength=(0.62, 0.645, 0.67), modifiers="sunz_corrected",
                          resolution=500)
        c02 = make_xarray("satellite_azimuth_angle", None)
        c03 = make_xarray("satellite_zenith_angle", None)
        c04 = make_xarray("solar_azimuth_angle", None)
        c05 = make_xarray("solar_zenith_angle", None)
        res = ref_cor([c01], [c02, c03, c04, c05])

        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["wavelength"] == (0.62, 0.645, 0.67)
        assert res.attrs["modifiers"] == ("sunz_corrected", "rayleigh_corrected_crefl",)
        assert res.attrs["calibration"] == "reflectance"
        assert res.attrs["resolution"] == 500
        assert res.attrs["name"] == "1"
        assert res.attrs["platform_name"] == "EOS-Aqua"
        assert res.attrs["sensor"] == "modis"
        assert res.attrs["units"] == "%"
        assert res.attrs["start_time"] == dt.datetime(2012, 8, 13, 18, 46, 1, 439838)
        assert res.attrs["end_time"] == dt.datetime(2012, 8, 13, 18, 57, 47, 746296)
        assert res.attrs["area"] == area
        assert res.attrs["ancillary_variables"] == []
        data = res.values
        assert abs(np.mean(data) - 52.09372623964498) < 1e-6
        assert data.shape == (3, 5)
        unique = np.unique(data)
        np.testing.assert_allclose(unique, [25.43670075, 52.93221561, 77.91226236])

    def test_reflectance_corrector_bad_prereqs(self):
        """Test ReflectanceCorrector modifier with wrong number of inputs."""
        from satpy.modifiers._crefl import ReflectanceCorrector
        ref_cor = ReflectanceCorrector("test")
        pytest.raises(ValueError, ref_cor, [1], [2, 3, 4], match="Not sure how to handle provided dependencies..*")
        pytest.raises(ValueError, ref_cor, [1, 2, 3, 4], [], match="Not sure how to handle provided dependencies..*")
        pytest.raises(ValueError, ref_cor, [], [1, 2, 3, 4], match="Not sure how to handle provided dependencies..*")

    @pytest.mark.parametrize(
        ("url", "dem_mock_cm", "dem_sds"),
        [
            (None, mock_cmgdem, "average elevation"),
            ("CMGDEM.hdf", mock_cmgdem, "averaged elevation"),
            ("tbase.hdf", mock_tbase, "Elevation"),
        ])
    def test_reflectance_corrector_different_chunks(self, tmpdir, url, dem_mock_cm, dem_sds):
        """Test that the modifier works with different chunk sizes for inputs.

        The modifier uses dask's "map_blocks". If the input chunks aren't the
        same an error is raised.

        """
        from satpy.modifiers._crefl import ReflectanceCorrector
        from satpy.tests.utils import make_dsq

        ref_cor = ReflectanceCorrector(
            optional_prerequisites=[
                make_dsq(name="satellite_azimuth_angle"),
                make_dsq(name="satellite_zenith_angle"),
                make_dsq(name="solar_azimuth_angle"),
                make_dsq(name="solar_zenith_angle")
            ],
            name="I01",
            prerequisites=[],
            wavelength=(0.6, 0.64, 0.68),
            resolution=371,
            calibration="reflectance",
            modifiers=("sunz_corrected_iband", "rayleigh_corrected_crefl_iband"),
            sensor="viirs",
            url=url,
            dem_sds=dem_sds,
        )

        area, data = self.data_area_ref_corrector()
        c01 = _make_viirs_xarray(data, area, "I01", "toa_bidirectional_reflectance",
                                 wavelength=(0.6, 0.64, 0.68), units="%",
                                 calibration="reflectance")
        c02 = _make_viirs_xarray(data, area, "satellite_azimuth_angle", "sensor_azimuth_angle")
        c02.data = c02.data.rechunk((1, -1))
        c03 = _make_viirs_xarray(data, area, "satellite_zenith_angle", "sensor_zenith_angle")
        c04 = _make_viirs_xarray(data, area, "solar_azimuth_angle", "solar_azimuth_angle")
        c05 = _make_viirs_xarray(data, area, "solar_zenith_angle", "solar_zenith_angle")

        with dem_mock_cm(tmpdir, url):
            res = ref_cor([c01], [c02, c03, c04, c05])

        # make sure it can actually compute
        res.compute()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010-2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Modis level 1b hdf-eos format reader.

Introduction
------------

The ``modis_l1b`` reader reads and calibrates Modis L1 image data in hdf-eos format. Files often have
a pattern similar to the following one:

.. parsed-literal::

    M[O/Y]D02[1/H/Q]KM.A[date].[time].[collection].[processing_time].hdf

Other patterns where "collection" and/or "proccessing_time" are missing might also work
(see the readers yaml file for details). Geolocation files (MOD03) are also supported.
The IMAPP direct broadcast naming format is also supported with names like:
``a1.12226.1846.1000m.hdf``.

Saturation Handling
-------------------

Band 2 of the MODIS sensor is available in 250m, 500m, and 1km resolutions.
The band data may include a special fill value to indicate when the detector
was saturated in the 250m version of the data. When the data is aggregated to
coarser resolutions this saturation fill value is converted to a
"can't aggregate" fill value. By default, Satpy will replace these fill values
with NaN to indicate they are invalid. This is typically undesired when
generating images for the data as they appear as "holes" in bright clouds.
To control this the keyword argument ``mask_saturated`` can be passed and set
to ``False`` to set these two fill values to the maximum valid value.

.. code-block:: python

    scene = satpy.Scene(filenames=filenames,
                        reader='modis_l1b',
                        reader_kwargs={'mask_saturated': False})
    scene.load(['2'])

Note that the saturation fill value can appear in other bands (ex. bands 7-19)
in addition to band 2. Also, the "can't aggregate" fill value is a generic
"catch all" for any problems encountered when aggregating high resolution bands
to lower resolutions. Filling this with the max valid value could replace
non-saturated invalid pixels with valid values.

Geolocation files
-----------------

For the 1km data (mod021km) geolocation files (mod03) are optional. If not given to the reader
1km geolocations will be interpolated from the 5km geolocation contained within the file.

For the 500m and 250m data geolocation files are needed.


References:
    - Modis gelocation description: http://www.icare.univ-lille1.fr/wiki/index.php/MODIS_geolocation
"""
import logging

import numpy as np
import xarray as xr

from satpy.readers.core.hdf4 import from_sds
from satpy.readers.core.hdfeos import HDFEOSBaseFileReader, HDFEOSGeoReader

logger = logging.getLogger(__name__)


class HDFEOSBandReader(HDFEOSBaseFileReader):
    """Handler for the regular band channels."""

    res = {"1": 1000,
           "Q": 250,
           "H": 500}

    res_to_possible_variable_names = {
        1000: ["EV_250_Aggr1km_RefSB",
               "EV_500_Aggr1km_RefSB",
               "EV_1KM_RefSB",
               "EV_1KM_Emissive"],
        500: ["EV_250_Aggr500_RefSB",
              "EV_500_RefSB"],
        250: ["EV_250_RefSB"],
    }

    def __init__(self, filename, filename_info, filetype_info, mask_saturated=True, **kwargs):
        """Init the file handler."""
        super().__init__(filename, filename_info, filetype_info, **kwargs)
        self._mask_saturated = mask_saturated

        ds = self.metadata["INVENTORYMETADATA"][
            "COLLECTIONDESCRIPTIONCLASS"]["SHORTNAME"]["VALUE"]
        self.resolution = self.res[ds[-3]]

    def get_dataset(self, key, info):
        """Read data from file and return the corresponding projectables."""
        if self.resolution != key["resolution"]:
            return
        var_name, band_index = self._get_band_variable_name_and_index(key["name"])
        subdata = self.sd.select(var_name)
        var_attrs = subdata.attributes()
        uncertainty = self.sd.select(var_name + "_Uncert_Indexes")
        chunks = self._chunks_for_variable(subdata)
        array = xr.DataArray(from_sds(subdata, self.filename, chunks=chunks)[band_index, :, :],
                             dims=["y", "x"]).astype(np.float32)
        valid_range = var_attrs["valid_range"]
        valid_min = np.float32(valid_range[0])
        valid_max = np.float32(valid_range[1])
        if not self._mask_saturated:
            array = self._fill_saturated(array, valid_max)
        array = self._mask_invalid(array, valid_min, valid_max)
        array = self._mask_uncertain_pixels(array, uncertainty, band_index)
        projectable = self._calibrate_data(key, info, array, var_attrs, band_index)

        # if ((platform_name == 'Aqua' and key['name'] in ["6", "27", "36"]) or
        #         (platform_name == 'Terra' and key['name'] in ["29"])):
        #     height, width = projectable.shape
        #     row_indices = projectable.mask.sum(1) == width
        #     if row_indices.sum() != height:
        #         projectable.mask[row_indices, :] = True

        # Get the orbit number
        # if not satscene.orbit:
        #     mda = self.data.attributes()["CoreMetadata.0"]
        #     orbit_idx = mda.index("ORBITNUMBER")
        #     satscene.orbit = mda[orbit_idx + 111:orbit_idx + 116]

        # Trimming out dead sensor lines (detectors) on terra:
        # (in addition channel 27, 30, 34, 35, and 36 are nosiy)
        # if satscene.satname == "terra":
        #     for band in ["29"]:
        #         if not satscene[band].is_loaded() or satscene[band].data.mask.all():
        #             continue
        #         width = satscene[band].data.shape[1]
        #         height = satscene[band].data.shape[0]
        #         indices = satscene[band].data.mask.sum(1) < width
        #         if indices.sum() == height:
        #             continue
        #         satscene[band] = satscene[band].data[indices, :]
        #         satscene[band].area = geometry.SwathDefinition(
        #             lons=satscene[band].area.lons[indices, :],
        #             lats=satscene[band].area.lats[indices, :])
        self._add_satpy_metadata(key, projectable)
        return projectable

    def _get_band_variable_name_and_index(self, band_name):
        variable_names = self.res_to_possible_variable_names[self.resolution]
        for variable_name in variable_names:
            subdata = self.sd.select(variable_name)
            var_attrs = subdata.attributes()
            try:
                band_index = self._get_band_index(var_attrs, band_name)
            except ValueError:
                # can't find band in list of bands
                continue
            return variable_name, band_index

    def _get_band_index(self, var_attrs, band_name):
        """Get the relative indices of the desired channel."""
        band_names = var_attrs["band_names"].split(",")
        index = band_names.index(band_name)
        return index

    def _fill_saturated(self, array, valid_max):
        """Replace saturation-related values with max reflectance.

        If the file handler was created with ``mask_saturated`` set to
        ``True`` then all invalid/fill values are set to NaN. If ``False``
        then the fill values 65528 and 65533 are set to the maximum valid
        value. These values correspond to "can't aggregate" and "saturation".

        Fill values:

        * 65535 Fill Value (includes reflective band data at night mode
          and completely missing L1A scans)
        * 65534 L1A DN is missing within a scan
        * 65533 Detector is saturated
        * 65532 Cannot compute zero point DN, e.g., SV is saturated
        * 65531 Detector is dead (see comments below)
        * 65530 RSB dn** below the minimum of the scaling range
        * 65529 TEB radiance or RSB dn exceeds the maximum of the scaling range
        * 65528 Aggregation algorithm failure
        * 65527 Rotation of Earth view Sector from nominal science collection position
        * 65526 Calibration coefficient b1 could not be computed
        * 65525 Subframe is dead
        * 65524 Both sides of the PCLW electronics on simultaneously
        * 65501 - 65523 (reserved for future use)
        * 65500 NAD closed upper limit

        """
        return array.where((array != 65533) & (array != 65528), valid_max)

    def _mask_invalid(self, array, valid_min, valid_max):
        """Replace fill values with NaN."""
        return array.where((array >= valid_min) & (array <= valid_max))

    def _mask_uncertain_pixels(self, array, uncertainty, band_index):
        if not self._mask_saturated:
            return array
        uncertainty_chunks = self._chunks_for_variable(uncertainty)
        band_uncertainty = from_sds(uncertainty, self.filename, chunks=uncertainty_chunks)[band_index, :, :]
        array = array.where(band_uncertainty < 15)
        return array

    def _calibrate_data(self, key, info, array, var_attrs, index):
        if key["calibration"] == "brightness_temperature":
            projectable = calibrate_bt(array, var_attrs, index, key["name"])
            info.setdefault("units", "K")
            info.setdefault("standard_name", "toa_brightness_temperature")
        elif key["calibration"] == "reflectance":
            projectable = calibrate_refl(array, var_attrs, index)
            info.setdefault("units", "%")
            info.setdefault("standard_name",
                            "toa_bidirectional_reflectance")
        elif key["calibration"] == "radiance":
            projectable = calibrate_radiance(array, var_attrs, index)
            info.setdefault("units", var_attrs.get("radiance_units"))
            info.setdefault("standard_name",
                            "toa_outgoing_radiance_per_unit_wavelength")
        elif key["calibration"] == "counts":
            projectable = calibrate_counts(array, var_attrs, index)
            info.setdefault("units", "counts")
            info.setdefault("standard_name", "counts")  # made up
        else:
            raise ValueError("Unknown calibration for "
                             "key: {}".format(key))
        projectable.attrs = info
        return projectable


class MixedHDFEOSReader(HDFEOSGeoReader, HDFEOSBandReader):
    """A file handler for the files that have both regular bands and geographical information in them."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Init the file handler."""
        HDFEOSGeoReader.__init__(self, filename, filename_info, filetype_info, **kwargs)
        HDFEOSBandReader.__init__(self, filename, filename_info, filetype_info, **kwargs)

    def get_dataset(self, key, info):
        """Get the dataset."""
        if key["name"] in HDFEOSGeoReader.DATASET_NAMES:
            return HDFEOSGeoReader.get_dataset(self, key, info)
        return HDFEOSBandReader.get_dataset(self, key, info)


def calibrate_counts(array, attributes, index):
    """Calibration for counts channels."""
    offset = np.float32(attributes["corrected_counts_offsets"][index])
    scale = np.float32(attributes["corrected_counts_scales"][index])
    array = (array - offset) * scale
    return array


def calibrate_radiance(array, attributes, index):
    """Calibration for radiance channels."""
    offset = np.float32(attributes["radiance_offsets"][index])
    scale = np.float32(attributes["radiance_scales"][index])
    array = (array - offset) * scale
    return array


def calibrate_refl(array, attributes, index):
    """Calibration for reflective channels."""
    offset = np.float32(attributes["reflectance_offsets"][index])
    scale = np.float32(attributes["reflectance_scales"][index])
    # convert to reflectance and convert from 1 to %
    array = (array - offset)
    array = array * (scale * 100)  # avoid extra dask tasks by combining scalars
    return array


def calibrate_bt(array, attributes, index, band_name):
    """Calibration for the emissive channels."""
    offset = np.float32(attributes["radiance_offsets"][index])
    scale = np.float32(attributes["radiance_scales"][index])

    array = (array - offset) * scale

    # Planck constant (Joule second)
    h__ = np.float32(6.6260755e-34)

    # Speed of light in vacuum (meters per second)
    c__ = np.float32(2.9979246e+8)

    # Boltzmann constant (Joules per Kelvin)
    k__ = np.float32(1.380658e-23)

    # Derived constants
    c_1 = 2 * h__ * c__ * c__
    c_2 = (h__ * c__) / k__

    # Effective central wavenumber (inverse centimeters)
    cwn = np.array([
        2.641775E+3, 2.505277E+3, 2.518028E+3, 2.465428E+3,
        2.235815E+3, 2.200346E+3, 1.477967E+3, 1.362737E+3,
        1.173190E+3, 1.027715E+3, 9.080884E+2, 8.315399E+2,
        7.483394E+2, 7.308963E+2, 7.188681E+2, 7.045367E+2],
        dtype=np.float32)

    # Temperature correction slope (no units)
    tcs = np.array([
        9.993411E-1, 9.998646E-1, 9.998584E-1, 9.998682E-1,
        9.998819E-1, 9.998845E-1, 9.994877E-1, 9.994918E-1,
        9.995495E-1, 9.997398E-1, 9.995608E-1, 9.997256E-1,
        9.999160E-1, 9.999167E-1, 9.999191E-1, 9.999281E-1],
        dtype=np.float32)

    # Temperature correction intercept (Kelvin)
    tci = np.array([
        4.770532E-1, 9.262664E-2, 9.757996E-2, 8.929242E-2,
        7.310901E-2, 7.060415E-2, 2.204921E-1, 2.046087E-1,
        1.599191E-1, 8.253401E-2, 1.302699E-1, 7.181833E-2,
        1.972608E-2, 1.913568E-2, 1.817817E-2, 1.583042E-2],
        dtype=np.float32)

    # Transfer wavenumber [cm^(-1)] to wavelength [m]
    cwn = 1. / (cwn * 100)

    # Some versions of the modis files do not contain all the bands.
    emmissive_channels = ["20", "21", "22", "23", "24", "25", "27", "28", "29",
                          "30", "31", "32", "33", "34", "35", "36"]
    global_index = emmissive_channels.index(band_name)

    cwn = cwn[global_index]
    tcs = tcs[global_index]
    tci = tci[global_index]
    array = c_2 / (cwn * np.log(c_1 / (1000000 * array * cwn ** 5) + 1))
    array = (array - tci) / tcs
    return array



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2023- Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""The multiple_sensors_isccpng_l1g_nc reader tests package.

This version tests the readers for ISCCP L1G data.

"""


import datetime as dt

import numpy as np
import pytest
import xarray as xr


@pytest.fixture
def nc_filename(tmp_path):
    """Create nc test data file and return its filename."""
    now = dt.datetime.now(dt.timezone.utc)
    filename = f"ISCCP-NG_L1g_demo_v5_res_0_05deg__temp_11_00um__{now:%Y%m%dT%H%M}.nc"
    filename_str = str(tmp_path / filename)

    jan_1970 = dt.datetime(1970, 1, 1, tzinfo=dt.timezone.utc)
    delta_t = now - jan_1970
    stime = delta_t.seconds
    etime = delta_t.seconds + 600
    # Create test data
    nscn = 3600
    npix = 7200
    lats = np.linspace(-90, 90, nscn)
    lons = np.linspace(-180, 180, npix)
    array = 27000 * np.ones((1, 3, nscn, npix))
    ds = xr.Dataset({"temp_11_00um": (("time", "layer", "latitude", "longitude"), array),
                     },
                    coords={"start_time": ("time", [stime]),
                            "end_time": ("time", [etime]),
                            "latitude": lats[:],
                            "longitude": lons[:]},
                    attrs={"scale_factor": 0.01, "units": "K"})

    ds["temp_11_00um"].attrs["_FillValue"] = -32767
    ds["temp_11_00um"].attrs["scale_factor"] = 0.01
    ds["temp_11_00um"].attrs["units"] = "K"
    ds["longitude"].attrs["standard_name"] = "longitude"
    ds["latitude"].attrs["standard_name"] = "latitude"
    ds["temp_11_00um"].attrs["standard_name"] = "temp_11_00um"
    comp = dict(zlib=True, complevel=5)
    encoding = {var: comp for var in ds.data_vars}
    ds.to_netcdf(filename_str, encoding=encoding)
    return filename_str


class TestISCCPNGL1gReader:
    """Test the IsccpngL1gFileHandler reader."""

    def test_read_isccpng_l1g(self, nc_filename):
        """Test reading reflectances and BT."""
        from satpy.scene import Scene

        # Read data
        scn_ = Scene(
            reader="multiple_sensors_isccpng_l1g_nc",
            filenames=[nc_filename])
        scn_.load(["temp_11_00um", "lon", "lat"])
        assert (scn_["lat"].shape == (3600, 7200))
        assert (scn_["lon"].shape == (3600, 7200))
        assert (scn_["temp_11_00um"].shape == (3600, 7200))
        assert (scn_["temp_11_00um"].values[0, 0] == 270)



"""Module containing CMSAF CLAAS v2 FileHandler."""

import datetime

from satpy.area import get_area_def
from satpy.readers.core.netcdf import NetCDF4FileHandler


def _is_georef_offset_present(date):
    # Reference: Product User Manual, section 3.
    # https://doi.org/10.5676/EUM_SAF_CM/CLAAS/V002_01
    return date < datetime.date(2017, 12, 6)


def _adjust_area_to_match_shifted_data(area):
    # Reference:
    # https://github.com/pytroll/satpy/wiki/SEVIRI-georeferencing-offset-correction
    offset = area.pixel_size_x / 2
    llx, lly, urx, ury = area.area_extent
    new_extent = [llx + offset, lly - offset, urx + offset, ury - offset]
    return area.copy(area_extent=new_extent)


FULL_DISK = get_area_def("msg_seviri_fes_3km")
FULL_DISK_WITH_OFFSET = _adjust_area_to_match_shifted_data(FULL_DISK)


class CLAAS2(NetCDF4FileHandler):
    """Handle CMSAF CLAAS-2 files."""

    def __init__(self, *args, **kwargs):
        """Initialise class."""
        super().__init__(*args, **kwargs, cache_handle=False,
                         auto_maskandscale=True)
        self.grid_size = self.file_content["/dimension/y"]

    @property
    def start_time(self):
        """Get start time from file."""
        # datetime module can't handle timezone identifier
        return datetime.datetime.fromisoformat(
                self["/attr/time_coverage_start"].rstrip("Z"))

    @property
    def end_time(self):
        """Get end time from file."""
        return datetime.datetime.fromisoformat(
                self["/attr/time_coverage_end"].rstrip("Z"))

    def available_datasets(self, configured_datasets=None):
        """Yield a collection of available datasets.

        Return a generator that will yield the datasets available in the loaded
        files.  See docstring in parent class for specification details.
        """
        # this method should work for any (CF-conform) NetCDF file, should it
        # be somewhere more generically available?  Perhaps in the
        # `NetCDF4FileHandler`?

        yield from super().available_datasets(configured_datasets)
        data_vars = [k for k in self.file_content
                     if k + "/dimensions" in self.file_content]
        for k in data_vars:
            # if it doesn't have a y-dimension we're not interested
            if "y" not in self.file_content[k + "/dimensions"]:
                continue
            ds_info = self._get_dsinfo(k)
            yield (True, ds_info)

    def _get_dsinfo(self, var):
        """Get metadata for variable.

        Return metadata dictionary for variable ``var``.
        """
        ds_info = {"name": var,
                   "file_type": self.filetype_info["file_type"]}
        # attributes for this data variable
        attrs = {k[len(f"{k:s}/attr")+1]: v
                 for (k, v) in self.file_content.items()
                 if k.startswith(f"{k:s}/attr")}
        # we don't need "special" attributes in our metadata here
        for unkey in {"_FillValue", "add_offset", "scale_factor"}:
            attrs.pop(unkey, None)
        return ds_info

    def get_dataset(self, dataset_id, info):
        """Get the dataset."""
        ds = self[dataset_id["name"]]
        if "time" in ds.dims:
            return ds.squeeze(["time"])

        return ds

    def get_area_def(self, dataset_id):
        """Get the area definition."""
        return self._get_subset_of_full_disk()

    def _get_subset_of_full_disk(self):
        """Get subset of the full disk.

        CLAAS2 products are provided on a grid that is slightly smaller
        than the full disk (excludes most of the space pixels).
        """
        full_disk = self._get_full_disk()
        offset = int((full_disk.width - self.grid_size) // 2)
        if offset > 0:
            return full_disk[offset:-offset, offset:-offset]
        return full_disk

    def _get_full_disk(self):
        if _is_georef_offset_present(self.start_time.date()):
            return FULL_DISK_WITH_OFFSET
        return FULL_DISK



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The glm_l2 reader tests package."""

import os
import unittest
from unittest import mock

import numpy as np
import xarray as xr


def setup_fake_dataset():
    """Create a fake dataset to avoid opening a file."""
    # flash_extent_density
    fed = (np.arange(10.).reshape((2, 5)) + 1.) * 50.
    fed = (fed + 1.) / 0.5
    fed = fed.astype(np.int16)
    fed = xr.DataArray(
        fed,
        dims=("y", "x"),
        attrs={
            "scale_factor": 0.5,
            "add_offset": -1.,
            "_FillValue": 0,
            "units": "Count per nominal    3136 microradian^2 pixel per 1.0 min",
            "grid_mapping": "goes_imager_projection",
            "standard_name": "flash_extent_density",
            "long_name": "Flash extent density",
        }
    )
    dqf = xr.DataArray(
        fed.data.copy().astype(np.uint8),
        dims=("y", "x"),
        attrs={
            "_FillValue": -1,
            "units": "1",
            "grid_mapping": "goes_imager_projection",
            "standard_name": "status_flag",
            "long_name": "GLM data quality flags",
            "flag_meanings": "valid invalid",
        }
    )
    # create a variable that won't be configured to test available_datasets
    not_configured = xr.DataArray(
        fed.data.copy(),
        dims=("y", "x"),
        attrs={
            "scale_factor": 0.5,
            "add_offset": -1.,
            "_FillValue": 0,
            "units": "1",
            "grid_mapping": "goes_imager_projection",
            "standard_name": "test",
            "long_name": "Test",
        }
    )
    x__ = xr.DataArray(
        range(5),
        attrs={"scale_factor": 2., "add_offset": -1.},
        dims=("x",),
    )
    y__ = xr.DataArray(
        range(2),
        attrs={"scale_factor": -2., "add_offset": 1.},
        dims=("y",),
    )
    proj = xr.DataArray(
        [],
        attrs={
            "semi_major_axis": 1.,
            "semi_minor_axis": 1.,
            "perspective_point_height": 1.,
            "longitude_of_projection_origin": -90.,
            "latitude_of_projection_origin": 0.,
            "sweep_angle_axis": u"x"
        }
    )
    fake_dataset = xr.Dataset(
        data_vars={
            "flash_extent_density": fed,
            "not_configured": not_configured,
            "DQF": dqf,
            "x": x__,
            "y": y__,
            "goes_imager_projection": proj,
            "nominal_satellite_subpoint_lat": np.array(0.0),
            "nominal_satellite_subpoint_lon": np.array(-89.5),
            "nominal_satellite_height": np.array(35786.02)
        },
        attrs={
            "time_coverage_start": "2017-09-20T17:30:40Z",
            "time_coverage_end": "2017-09-20T17:41:17Z",
            "spatial_resolution": "2km at nadir",
        }
    )
    return fake_dataset


class TestGLML2FileHandler(unittest.TestCase):
    """Tests for the GLM L2 reader."""

    @mock.patch("satpy.readers.core.abi.xr")
    def setUp(self, xr_):
        """Create a fake file handler to test."""
        from satpy.readers.glm_l2 import NCGriddedGLML2
        fake_dataset = setup_fake_dataset()
        xr_.open_dataset.return_value = fake_dataset
        self.reader = NCGriddedGLML2("filename",
                                     {"platform_shortname": "G16",
                                      "scene_abbr": "C", "scan_mode": "M3"},
                                     {"filetype": "glm_l2_imagery"})

    def test_basic_attributes(self):
        """Test getting basic file attributes."""
        import datetime as dt
        assert self.reader.start_time == dt.datetime(2017, 9, 20, 17, 30, 40)
        assert self.reader.end_time == dt.datetime(2017, 9, 20, 17, 41, 17)

    def test_get_dataset(self):
        """Test the get_dataset method."""
        from satpy.tests.utils import make_dataid
        key = make_dataid(name="flash_extent_density")
        res = self.reader.get_dataset(key, {"info": "info"})
        exp = {"instrument_ID": None,
               "modifiers": (),
               "name": "flash_extent_density",
               "orbital_parameters": {"projection_altitude": 1.0,
                                      "projection_latitude": 0.0,
                                      "projection_longitude": -90.0,
                                      # 'satellite_nominal_altitude': 35786.02,
                                      "satellite_nominal_latitude": 0.0,
                                      "satellite_nominal_longitude": -89.5},
               "orbital_slot": None,
               "platform_name": "GOES-16",
               "platform_shortname": "G16",
               "production_site": None,
               "scan_mode": "M3",
               "scene_abbr": "C",
               "scene_id": None,
               "spatial_resolution": "2km at nadir",
               "sensor": "glm",
               "timeline_ID": None,
               "grid_mapping": "goes_imager_projection",
               "standard_name": "flash_extent_density",
               "long_name": "Flash extent density",
               "units": "Count per nominal    3136 microradian^2 pixel per 1.0 min"}

        assert res.attrs == exp

    def test_get_dataset_dqf(self):
        """Test the get_dataset method with special DQF var."""
        from satpy.tests.utils import make_dataid
        key = make_dataid(name="DQF")
        res = self.reader.get_dataset(key, {"info": "info"})
        exp = {"instrument_ID": None,
               "modifiers": (),
               "name": "DQF",
               "orbital_parameters": {"projection_altitude": 1.0,
                                      "projection_latitude": 0.0,
                                      "projection_longitude": -90.0,
                                      # 'satellite_nominal_altitude': 35786.02,
                                      "satellite_nominal_latitude": 0.0,
                                      "satellite_nominal_longitude": -89.5},
               "orbital_slot": None,
               "platform_name": "GOES-16",
               "platform_shortname": "G16",
               "production_site": None,
               "scan_mode": "M3",
               "scene_abbr": "C",
               "scene_id": None,
               "spatial_resolution": "2km at nadir",
               "sensor": "glm",
               "timeline_ID": None,
               "grid_mapping": "goes_imager_projection",
               "units": "1",
               "_FillValue": -1,
               "standard_name": "status_flag",
               "long_name": "GLM data quality flags",
               "flag_meanings": "valid invalid"}

        assert res.attrs == exp
        assert np.issubdtype(res.dtype, np.integer)


class TestGLML2Reader(unittest.TestCase):
    """Test high-level reading functionality of GLM L2 reader."""

    yaml_file = "glm_l2.yaml"

    @mock.patch("satpy.readers.core.abi.xr")
    def setUp(self, xr_):
        """Create a fake reader to test."""
        from satpy._config import config_search_paths
        from satpy.readers.core.loading import load_reader
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        fake_dataset = setup_fake_dataset()
        xr_.open_dataset.return_value = fake_dataset
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "OR_GLM-L2-GLMC-M3_G16_s20192862159000_e20192862200000_c20192862200350.nc",
            "CSPP_CG_GLM-L2-GLMC-M3_G16_s20192862159000_e20192862200000_c20192862200350.nc",
        ])
        assert len(loadables) == 2
        r.create_filehandlers(loadables)
        self.reader = r

    def test_available_datasets(self):
        """Test that resolution is added to YAML configured variables."""
        # make sure we have some files
        assert self.reader.file_handlers
        available_datasets = list(self.reader.available_dataset_ids)
        # flash_extent_density, DQF, and not_configured are available in our tests
        assert len(available_datasets) == 3
        for ds_id in available_datasets:
            assert ds_id["resolution"] == 2000
        # make sure not_configured was discovered
        names = [dataid["name"] for dataid in available_datasets]
        assert "not_configured" in names



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

r"""SEVIRI Level 1.5 native format reader.

Introduction
____________
The ``seviri_l1b_native`` reader reads and calibrates MSG-SEVIRI L1.5 image data in binary format. The format is
explained in the `MSG Level 1.5 Native Format File Definition`_. The files are usually named as
follows:

.. code-block:: none

    MSG4-SEVI-MSG15-0100-NA-20210302124244.185000000Z-NA.nat

Reader Arguments
----------------
Some arguments can be provided to the reader to change its behaviour. These are
provided through the `Scene` instantiation, eg::

  scn = Scene(filenames=filenames, reader="seviri_l1b_native", reader_kwargs={'fill_disk': True})

To see the full list of arguments that can be provided, look into the documentation
of :class:`NativeMSGFileHandler`.

Example:
--------
Here is an example how to read the data in satpy.

NOTE: When loading the data, the orientation
of the image can be set with ``upper_right_corner``-keyword.
Possible options are ``NW``, ``NE``, ``SW``, ``SE``, or ``native``.

.. code-block:: python

    from satpy import Scene

    filenames = ['MSG4-SEVI-MSG15-0100-NA-20210302124244.185000000Z-NA.nat']
    scn = Scene(filenames=filenames, reader='seviri_l1b_native')
    scn.load(['VIS006', 'IR_108'], upper_right_corner='NE')
    print(scn['IR_108'])


Output:

.. code-block:: none

    <xarray.DataArray 'reshape-969ef97d34b7b0c70ca19f53c6abcb68' (y: 3712, x: 3712)>
    dask.array<truediv, shape=(3712, 3712), dtype=float32, chunksize=(928, 3712), chunktype=numpy.ndarray>
    Coordinates:
        acq_time  (y) datetime64[ns] NaT NaT NaT NaT NaT NaT ... NaT NaT NaT NaT NaT
        crs       object PROJCRS["unknown",BASEGEOGCRS["unknown",DATUM["unknown",...
      * y         (y) float64 -5.566e+06 -5.563e+06 ... 5.566e+06 5.569e+06
      * x         (x) float64 5.566e+06 5.563e+06 5.56e+06 ... -5.566e+06 -5.569e+06
    Attributes:
        orbital_parameters:       {'projection_longitude': 0.0, 'projection_latit...
        time_parameters:          {'nominal_start_time': datetime.datetime(2021, ...
        units:                    K
        wavelength:               10.8 µm (9.8-11.8 µm)
        standard_name:            toa_brightness_temperature
        platform_name:            Meteosat-11
        sensor:                   seviri
        georef_offset_corrected:  True
        start_time:               2021-03-02 12:30:11.584603
        end_time:                 2021-03-02 12:45:09.949762
        reader:                   seviri_l1b_native
        area:                     Area ID: msg_seviri_fes_3km\\nDescription: MSG S...
        name:                     IR_108
        resolution:               3000.403165817
        calibration:              brightness_temperature
        modifiers:                ()
        _satpy_id:                DataID(name='IR_108', wavelength=WavelengthRang...
        ancillary_variables:      []


References:
    - `EUMETSAT Product Navigator`_
    - `MSG Level 1.5 Native Format File Definition`_

.. _EUMETSAT Product Navigator:
    https://navigator.eumetsat.int/product/EO:EUM:DAT:MSG:HRSEVIRI
.. _MSG Level 1.5 Native Format File Definition:
    https://user.eumetsat.int/s3/eup-strapi-media/pdf_fg15_msg_native_format_15_6b513c5bb3.pdf
"""

import datetime as dt
import logging
import warnings

import dask.array as da
import numpy as np
import xarray as xr
from pyresample import geometry

from satpy._compat import cached_property
from satpy.readers.core._geos_area import get_area_definition, get_geos_area_naming
from satpy.readers.core.eum import get_service_mode, recarray2dict, time_cds_short
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.seviri import (
    CHANNEL_NAMES,
    HRV_NUM_COLUMNS,
    HRV_NUM_LINES,
    REPEAT_CYCLE_DURATION,
    SATNUM,
    VISIR_NUM_COLUMNS,
    VISIR_NUM_LINES,
    CalibParams,
    GsicsCoefficients,
    MeirinkCoefficients,
    NominalCoefficients,
    NoValidOrbitParams,
    OrbitPolynomialFinder,
    ScanParams,
    SEVIRICalibrationHandler,
    add_scanline_acq_time,
    calculate_area_extent,
    create_coef_dict,
    dec10216,
    get_cds_time,
    get_satpos,
    pad_data_horizontally,
    pad_data_vertically,
    round_nom_time,
)
from satpy.readers.core.utils import fromfile, generic_open, reduce_mda
from satpy.readers.seviri_l1b_native_hdr import (
    DEFAULT_15_SECONDARY_PRODUCT_HEADER,
    GSDTRecords,
    get_native_header,
    native_trailer,
)
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger("native_msg")
CHUNK_SIZE = get_legacy_chunk_size()
ASCII_STARTSWITH = b"FormatName                  : NATIVE"


class NativeMSGFileHandler(BaseFileHandler):
    """SEVIRI native format reader.

    **Calibration**

    See :mod:`satpy.readers.core.seviri`.

    **Padding channel data to full disk**

    By providing the `fill_disk` as True in the `reader_kwargs`, the channel is loaded
    as full disk, padded with no-data where necessary. This is especially useful for the
    HRV channel, but can also be used for RSS and ROI data. By default, the original,
    unpadded, data are loaded::

        scene = satpy.Scene(filenames,
                            reader='seviri_l1b_native',
                            reader_kwargs={'fill_disk': False})

    **Metadata**

    See :mod:`satpy.readers.core.seviri`.

    """

    def __init__(self, filename, filename_info, filetype_info,
                 calib_mode="nominal", fill_disk=False, ext_calib_coefs=None,
                 include_raw_metadata=False, mda_max_array_size=100):
        """Initialize the reader."""
        super(NativeMSGFileHandler, self).__init__(filename,
                                                   filename_info,
                                                   filetype_info)
        self.platform_name = None
        self.calib_mode = calib_mode
        self.ext_calib_coefs = ext_calib_coefs or {}
        self.fill_disk = fill_disk
        self.include_raw_metadata = include_raw_metadata
        self.mda_max_array_size = mda_max_array_size

        # Declare required variables.
        self.header = {}
        self.mda = {}
        self.trailer = {}

        # Read header, prepare dask-array, read trailer and initialize image boundaries
        # Available channels are known only after the header has been read
        self.header_type = get_native_header(has_archive_header(self.filename))
        self._read_header()
        self._make_dask_array_with_map_blocks()
        self._read_trailer()
        self.image_boundaries = ImageBoundaries(self.header, self.trailer, self.mda)

    def _make_dask_array_with_map_blocks(self):
        """Make the dask array using the ``da.map_blocks()`` functionality."""
        dtype = self._get_data_dtype()
        chunks = da.core.normalize_chunks(
            "auto",
            shape=(self.mda["number_of_lines"],),
            dtype=dtype)
        self._dask_array = da.map_blocks(
            _get_array,
            dtype=dtype,
            chunks=chunks,
            meta=np.array([], dtype=dtype),
            # The following will be passed as keyword arguments to the `_get_array()` function.
            filename=self.filename,
            hdr_size=self.header_type.itemsize
        )

    @property
    def _repeat_cycle_duration(self):
        """Get repeat cycle duration from the trailer."""
        if self.trailer["15TRAILER"]["ImageProductionStats"]["ActualScanningSummary"]["ReducedScan"] == 1:
            return 5
        return REPEAT_CYCLE_DURATION

    @property
    def nominal_start_time(self):
        """Get the repeat cycle nominal start time from file header and round it to expected nominal time slot."""
        tm = self.header["15_DATA_HEADER"]["ImageAcquisition"]["PlannedAcquisitionTime"]["TrueRepeatCycleStart"]
        return round_nom_time(tm, time_delta=dt.timedelta(minutes=self._repeat_cycle_duration))

    @property
    def nominal_end_time(self):
        """Get the repeat cycle nominal end time from file header and round it to expected nominal time slot."""
        tm = self.header["15_DATA_HEADER"]["ImageAcquisition"]["PlannedAcquisitionTime"]["PlannedRepeatCycleEnd"]
        return round_nom_time(tm, time_delta=dt.timedelta(minutes=self._repeat_cycle_duration))

    @property
    def observation_start_time(self):
        """Get observation start time from trailer."""
        return self.trailer["15TRAILER"]["ImageProductionStats"][
            "ActualScanningSummary"]["ForwardScanStart"]

    @property
    def observation_end_time(self):
        """Get observation end time from trailer."""
        return self.trailer["15TRAILER"]["ImageProductionStats"][
            "ActualScanningSummary"]["ForwardScanEnd"]

    @property
    def start_time(self):
        """Get general start time for this file."""
        return self.nominal_start_time

    @property
    def end_time(self):
        """Get the general end time for this file."""
        return self.nominal_end_time

    def _get_data_dtype(self):
        """Get the dtype of the file based on the actual available channels."""
        pkhrec = [
            ("GP_PK_HEADER", GSDTRecords.gp_pk_header),
            ("GP_PK_SH1", GSDTRecords.gp_pk_sh1)
        ]
        pk_head_dtype = np.dtype(pkhrec)

        def get_lrec(cols):
            lrec = [
                ("gp_pk", pk_head_dtype),
                ("version", np.uint8),
                ("satid", np.uint16),
                ("time", (np.uint16, 5)),
                ("lineno", np.uint32),
                ("chan_id", np.uint8),
                ("acq_time", time_cds_short),
                ("line_validity", np.uint8),
                ("line_rquality", np.uint8),
                ("line_gquality", np.uint8),
                ("line_data", (np.uint8, cols))
            ]

            return lrec

        # each pixel is 10-bits -> one line of data has 25% more bytes
        # than the number of columns suggest (10/8 = 1.25)
        visir_rec = get_lrec(int(self.mda["number_of_columns"] * 1.25))
        drec = [("visir", (visir_rec, self._number_of_visir_channels()))]

        if self.mda["available_channels"]["HRV"]:
            hrv_rec = get_lrec(int(self.mda["hrv_number_of_columns"] * 1.25))
            drec.append(("hrv", (hrv_rec, 3)))

        return np.dtype(drec)

    def _number_of_visir_channels(self):
        """Return the number of visir channels, i.e. all channels excluding ``HRV``."""
        return len([s for s in self.mda["channel_list"] if not s == "HRV"])

    def _read_header(self):
        """Read the header info."""
        self.header.update(read_header(self.filename))

        if "15_SECONDARY_PRODUCT_HEADER" not in self.header:
            # No archive header, that means we have a complete file
            # including all channels.
            self.header["15_SECONDARY_PRODUCT_HEADER"] = DEFAULT_15_SECONDARY_PRODUCT_HEADER

        data15hd = self.header["15_DATA_HEADER"]
        sec15hd = self.header["15_SECONDARY_PRODUCT_HEADER"]

        # Set the list of available channels:
        self.mda["available_channels"] = get_available_channels(self.header)
        self.mda["channel_list"] = [i for i in CHANNEL_NAMES.values()
                                    if self.mda["available_channels"][i]]

        self.platform_id = data15hd[
            "SatelliteStatus"]["SatelliteDefinition"]["SatelliteId"]
        self.mda["platform_name"] = "Meteosat-" + SATNUM[self.platform_id]
        self.mda["offset_corrected"] = data15hd["GeometricProcessing"][
                                           "EarthModel"]["TypeOfEarthModel"] == 2

        equator_radius = data15hd["GeometricProcessing"][
                             "EarthModel"]["EquatorialRadius"] * 1000.
        north_polar_radius = data15hd[
                                 "GeometricProcessing"]["EarthModel"]["NorthPolarRadius"] * 1000.
        south_polar_radius = data15hd[
                                 "GeometricProcessing"]["EarthModel"]["SouthPolarRadius"] * 1000.
        polar_radius = (north_polar_radius + south_polar_radius) * 0.5
        ssp_lon = data15hd["ImageDescription"][
            "ProjectionDescription"]["LongitudeOfSSP"]

        self.mda["projection_parameters"] = {"a": equator_radius,
                                             "b": polar_radius,
                                             "h": 35785831.00,
                                             "ssp_longitude": ssp_lon}

        north = int(sec15hd["NorthLineSelectedRectangle"]["Value"])
        east = int(sec15hd["EastColumnSelectedRectangle"]["Value"])
        south = int(sec15hd["SouthLineSelectedRectangle"]["Value"])
        west = int(sec15hd["WestColumnSelectedRectangle"]["Value"])

        ncolumns = west - east + 1
        nrows = north - south + 1

        # check if the file has less rows or columns than
        # the maximum, if so it is a rapid scanning service
        # or region of interest file
        if (nrows < VISIR_NUM_LINES) or (ncolumns < VISIR_NUM_COLUMNS):
            self.mda["is_full_disk"] = False
        else:
            self.mda["is_full_disk"] = True

        # If the number of columns in the file is not divisible by 4,
        # UMARF will add extra columns to the file
        modulo = ncolumns % 4
        padding = 0
        if modulo > 0:
            padding = 4 - modulo
        cols_visir = ncolumns + padding

        # Check the VISIR calculated column dimension against
        # the header information
        cols_visir_hdr = int(sec15hd["NumberColumnsVISIR"]["Value"])
        if cols_visir_hdr != cols_visir:
            logger.warning(
                "Number of VISIR columns from the header is incorrect!")
            logger.warning("Header: %d", cols_visir_hdr)
            logger.warning("Calculated: = %d", cols_visir)

        # HRV Channel - check if the area is reduced in east west
        # direction as this affects the number of columns in the file
        cols_hrv_hdr = int(sec15hd["NumberColumnsHRV"]["Value"])
        if ncolumns < VISIR_NUM_COLUMNS:
            cols_hrv = cols_hrv_hdr
        else:
            cols_hrv = int(cols_hrv_hdr / 2)

        # self.mda represents the 16bit dimensions not 10bit
        self.mda["number_of_lines"] = int(sec15hd["NumberLinesVISIR"]["Value"])
        self.mda["number_of_columns"] = cols_visir
        self.mda["hrv_number_of_lines"] = int(sec15hd["NumberLinesHRV"]["Value"])
        self.mda["hrv_number_of_columns"] = cols_hrv

        if "15_MAIN_PRODUCT_HEADER" not in self.header:
            logger.info("Quality flag check was not possible due to missing 15_MAIN_PRODUCT_HEADER.")
        elif self.header["15_MAIN_PRODUCT_HEADER"]["QQOV"]["Value"] == "NOK":
            warnings.warn(
                "The quality flag for this file indicates not OK. "
                "Use this data with caution!",
                UserWarning,
                stacklevel=2
            )

    def _read_trailer(self):

        hdr_size = self.header_type.itemsize
        data_size = (self._get_data_dtype().itemsize *
                     self.mda["number_of_lines"])

        data = fromfile(self.filename, dtype=native_trailer, count=1, offset=hdr_size + data_size)

        self.trailer.update(recarray2dict(data))

    def get_area_def(self, dataset_id):
        """Get the area definition of the band.

        In general, image data from one window/area is available. For the HRV channel in FES mode, however,
        data from two windows ('Lower' and 'Upper') are available. Hence, we collect lists of area-extents
        and corresponding number of image lines/columns. In case of FES HRV data, two area definitions are
        computed, stacked and squeezed. For other cases, the lists will only have one entry each, from which
        a single area definition is computed.

        Note that the AreaDefinition area extents returned by this function for Native data will be slightly
        different compared to the area extents returned by the SEVIRI HRIT reader.
        This is due to slightly different pixel size values when calculated using the data available in the files. E.g.
        for the 3 km grid:

        ``Native: data15hd['ImageDescription']['ReferenceGridVIS_IR']['ColumnDirGridStep'] == 3000.4031658172607``
        ``HRIT:                            np.deg2rad(2.**16 / pdict['lfac']) * pdict['h'] == 3000.4032785810186``

        This results in the Native 3 km full-disk area extents being approx. 20 cm shorter in each direction.

        The method for calculating the area extents used by the HRIT reader (CFAC/LFAC mechanism) keeps the
        highest level of numeric precision and is used as reference by EUM. For this reason, the standard area
        definitions defined in the `areas.yaml` file correspond to the HRIT ones.

        """
        pdict = dict()
        pdict["a"] = self.mda["projection_parameters"]["a"]
        pdict["b"] = self.mda["projection_parameters"]["b"]
        pdict["h"] = self.mda["projection_parameters"]["h"]
        pdict["ssp_lon"] = self.mda["projection_parameters"]["ssp_longitude"]

        area_naming_input_dict = {"platform_name": "msg",
                                  "instrument_name": "seviri",
                                  "resolution": int(dataset_id["resolution"])
                                  }
        area_naming = get_geos_area_naming({**area_naming_input_dict,
                                            **get_service_mode("seviri", pdict["ssp_lon"])})

        pdict["a_name"] = area_naming["area_id"]
        pdict["a_desc"] = area_naming["description"]
        pdict["p_id"] = ""

        area_extent = self.get_area_extent(dataset_id)
        areas = list()
        for aex, nlines, ncolumns in zip(area_extent["area_extent"], area_extent["nlines"], area_extent["ncolumns"]):
            pdict["nlines"] = nlines
            pdict["ncols"] = ncolumns
            areas.append(get_area_definition(pdict, aex))

        if len(areas) == 2:
            area = geometry.StackedAreaDefinition(areas[0], areas[1])
            area = area.squeeze()
        else:
            area = areas[0]

        return area

    def get_area_extent(self, dataset_id):
        """Get the area extent of the file.

        Until December 2017, the data is shifted by 1.5km SSP North and West against the nominal GEOS projection. Since
        December 2017 this offset has been corrected. A flag in the data indicates if the correction has been applied.
        If no correction was applied, adjust the area extent to match the shifted data.

        For more information see Section 3.1.4.2 in the MSG Level 1.5 Image Data Format Description. The correction
        of the area extent is documented in a `developer's memo <https://github.com/pytroll/satpy/wiki/
        SEVIRI-georeferencing-offset-correction>`_.
        """
        data15hd = self.header["15_DATA_HEADER"]

        # check for Earth model as this affects the north-south and
        # west-east offsets
        # section 3.1.4.2 of MSG Level 1.5 Image Data Format Description
        earth_model = data15hd["GeometricProcessing"]["EarthModel"][
            "TypeOfEarthModel"]
        if earth_model == 2:
            ns_offset = 0
            we_offset = 0
        elif earth_model == 1:
            ns_offset = -0.5
            we_offset = 0.5
            if dataset_id["name"] == "HRV":
                ns_offset = -1.5
                we_offset = 1.5
        else:
            raise NotImplementedError(
                "Unrecognised Earth model: {}".format(earth_model)
            )

        if dataset_id["name"] == "HRV":
            grid_origin = data15hd["ImageDescription"]["ReferenceGridHRV"]["GridOrigin"]
            center_point = (HRV_NUM_COLUMNS / 2) - 2
            column_step = data15hd["ImageDescription"]["ReferenceGridHRV"]["ColumnDirGridStep"] * 1000.0
            line_step = data15hd["ImageDescription"]["ReferenceGridHRV"]["LineDirGridStep"] * 1000.0
            nlines_fulldisk = HRV_NUM_LINES
            ncolumns_fulldisk = HRV_NUM_COLUMNS
        else:
            grid_origin = data15hd["ImageDescription"]["ReferenceGridVIS_IR"]["GridOrigin"]
            center_point = VISIR_NUM_COLUMNS / 2
            column_step = data15hd["ImageDescription"]["ReferenceGridVIS_IR"]["ColumnDirGridStep"] * 1000.0
            line_step = data15hd["ImageDescription"]["ReferenceGridVIS_IR"]["LineDirGridStep"] * 1000.0
            nlines_fulldisk = VISIR_NUM_LINES
            ncolumns_fulldisk = VISIR_NUM_COLUMNS

        # Calculations assume grid origin is south-east corner
        # section 7.2.4 of MSG Level 1.5 Image Data Format Description
        origins = {0: "NW", 1: "SW", 2: "SE", 3: "NE"}
        if grid_origin != 2:
            msg = "Grid origin not supported number: {}, {} corner".format(
                grid_origin, origins[grid_origin]
            )
            raise NotImplementedError(msg)

        aex_data = {"area_extent": [], "nlines": [], "ncolumns": []}

        img_bounds = self.image_boundaries.get_img_bounds(dataset_id, self.is_roi())
        for south_bound, north_bound, east_bound, west_bound in zip(*img_bounds.values()):

            if self.fill_disk:
                east_bound, west_bound = 1, ncolumns_fulldisk
                if not self.mda["is_full_disk"]:
                    south_bound, north_bound = 1, nlines_fulldisk

            nlines = north_bound - south_bound + 1
            ncolumns = west_bound - east_bound + 1

            area_dict = {"center_point": center_point,
                         "east": east_bound,
                         "west": west_bound,
                         "south": south_bound,
                         "north": north_bound,
                         "column_step": column_step,
                         "line_step": line_step,
                         "column_offset": we_offset,
                         "line_offset": ns_offset
                         }

            aex = calculate_area_extent(area_dict)

            aex_data["area_extent"].append(aex)
            aex_data["nlines"].append(nlines)
            aex_data["ncolumns"].append(ncolumns)

        return aex_data

    def is_roi(self):
        """Check if data covers a selected region of interest (ROI).

        Standard RSS data consists of 3712 columns and 1392 lines, covering the three northmost segments
        of the SEVIRI disk. Hence, if the data does not cover the full disk, nor the standard RSS region
        in RSS mode, it's assumed to be ROI data.
        """
        is_rapid_scan = self.trailer["15TRAILER"]["ImageProductionStats"]["ActualScanningSummary"]["ReducedScan"]

        # Standard RSS data is assumed to cover the three northmost segments, thus consisting of all 3712 columns and
        # the 1392 northmost lines
        nlines = int(self.mda["number_of_lines"])
        ncolumns = int(self.mda["number_of_columns"])
        north_bound = int(self.header["15_SECONDARY_PRODUCT_HEADER"]["NorthLineSelectedRectangle"]["Value"])

        is_top3segments = (ncolumns == VISIR_NUM_COLUMNS and nlines == 1392 and north_bound == VISIR_NUM_LINES)

        return not self.mda["is_full_disk"] and not (is_rapid_scan and is_top3segments)

    def get_dataset(self, dataset_id, dataset_info):
        """Get the dataset."""
        if dataset_id["name"] not in self.mda["channel_list"]:
            raise KeyError("Channel % s not available in the file" % dataset_id["name"])
        elif dataset_id["name"] not in ["HRV"]:
            data = self._get_visir_channel(dataset_id)
        else:
            data = self._get_hrv_channel()

        xarr = xr.DataArray(data, dims=["y", "x"]).where(data != 0).astype(np.float32)

        if xarr is None:
            return None

        dataset = self.calibrate(xarr, dataset_id)
        self._add_scanline_acq_time(dataset, dataset_id)
        self._update_attrs(dataset, dataset_info)

        if self.fill_disk and not (dataset_id["name"] != "HRV" and self.mda["is_full_disk"]):
            padder = Padder(dataset_id,
                            self.image_boundaries.get_img_bounds(dataset_id, self.is_roi()),
                            self.mda["is_full_disk"])
            dataset = padder.pad_data(dataset)

        return dataset

    def _get_visir_channel(self, dataset_id):
        shape = (self.mda["number_of_lines"], self.mda["number_of_columns"])
        # Check if there is only 1 channel in the list as a change
        # is needed in the array assignment ie channel id is not present
        if len(self.mda["channel_list"]) == 1:
            raw = self._dask_array["visir"]["line_data"]
        else:
            i = self.mda["channel_list"].index(dataset_id["name"])
            raw = self._dask_array["visir"]["line_data"][:, i, :]
        data = dec10216(raw.flatten())
        data = data.reshape(shape)
        return data

    def _get_hrv_channel(self):
        shape = (self.mda["hrv_number_of_lines"], self.mda["hrv_number_of_columns"])
        shape_layer = (self.mda["number_of_lines"], self.mda["hrv_number_of_columns"])

        data_list = []
        for i in range(3):
            raw = self._dask_array["hrv"]["line_data"][:, i, :]
            data = dec10216(raw.flatten())
            data = data.reshape(shape_layer)
            data_list.append(data)

        return np.stack(data_list, axis=1).reshape(shape)

    def calibrate(self, data, dataset_id):
        """Calibrate the data."""
        tic = dt.datetime.now()
        calib = self._get_calibration_handler(dataset_id)
        res = calib.calibrate(data, dataset_id["calibration"])
        logger.debug("Calibration time " + str(dt.datetime.now() - tic))
        return res

    def _get_calibration_handler(self, dataset_id):
        channel_name = dataset_id["name"]
        calib_params = CalibParams(
            mode=self.calib_mode.upper(),
            internal_coefs=self._get_calib_coefs(channel_name),
            external_coefs=self.ext_calib_coefs,
            radiance_type=self._get_radiance_type(channel_name)
        )
        scan_params = ScanParams(self.platform_id, channel_name,
                                 self.observation_start_time)
        return SEVIRICalibrationHandler(calib_params, scan_params)

    def _get_calib_coefs(self, channel_name):
        """Get coefficients for calibration from counts to radiance."""
        band_idx = self._get_band_index(channel_name)
        coefs_nominal = self.header["15_DATA_HEADER"][
            "RadiometricProcessing"]["Level15ImageCalibration"]
        coefs_gsics = self.header["15_DATA_HEADER"][
            "RadiometricProcessing"]["MPEFCalFeedback"]
        nominal_coefs = NominalCoefficients(
            channel_name, coefs_nominal["CalSlope"][band_idx], coefs_nominal["CalOffset"][band_idx]
        )
        gsics_coefs = GsicsCoefficients(
            channel_name, coefs_gsics["GSICSCalCoeff"][band_idx], coefs_gsics["GSICSOffsetCount"][band_idx]
        )
        meirink_coefs = MeirinkCoefficients(
            self.platform_id, channel_name, self.observation_start_time
        )
        return create_coef_dict(
            nominal_coefs,
            gsics_coefs,
            meirink_coefs
        )

    def _get_band_index(self, channel_name):
        # even though all the channels may not be present in the file,
        # the header does have calibration coefficients for all the channels
        # hence, this channel index needs to refer to full channel list
        return list(CHANNEL_NAMES.values()).index(channel_name)

    def _get_radiance_type(self, channel_name):
        band_idx = self._get_band_index(channel_name)
        radiance_types = self.header["15_DATA_HEADER"]["ImageDescription"][
            "Level15ImageProduction"]["PlannedChanProcessing"]
        return radiance_types[band_idx]

    def _add_scanline_acq_time(self, dataset, dataset_id):
        """Add scanline acquisition time to the given dataset."""
        if dataset_id["name"] == "HRV":
            tline = self._get_acq_time_hrv()
        else:
            tline = self._get_acq_time_visir(dataset_id)
        acq_time = get_cds_time(days=tline["Days"], msecs=tline["Milliseconds"])
        add_scanline_acq_time(dataset, acq_time)

    def _get_acq_time_hrv(self):
        """Get raw acquisition time for HRV channel."""
        tline = self._dask_array["hrv"]["acq_time"]
        tline0 = tline[:, 0]
        tline1 = tline[:, 1]
        tline2 = tline[:, 2]
        return da.stack((tline0, tline1, tline2), axis=1).reshape(
            self.mda["hrv_number_of_lines"]).compute()

    def _get_acq_time_visir(self, dataset_id):
        """Get raw acquisition time for VIS/IR channels."""
        # Check if there is only 1 channel in the list as a change
        # is needed in the array assignment, i.e. channel id is not present
        if len(self.mda["channel_list"]) == 1:
            return self._dask_array["visir"]["acq_time"].compute()
        i = self.mda["channel_list"].index(dataset_id["name"])
        return self._dask_array["visir"]["acq_time"][:, i].compute()

    def _update_attrs(self, dataset, dataset_info):
        """Update dataset attributes."""
        dataset.attrs["units"] = dataset_info["units"]
        dataset.attrs["wavelength"] = dataset_info["wavelength"]
        dataset.attrs["standard_name"] = dataset_info["standard_name"]
        dataset.attrs["platform_name"] = self.mda["platform_name"]
        dataset.attrs["sensor"] = "seviri"
        dataset.attrs["georef_offset_corrected"] = self.mda[
            "offset_corrected"]
        dataset.attrs["time_parameters"] = {
            "nominal_start_time": self.nominal_start_time,
            "nominal_end_time": self.nominal_end_time,
            "observation_start_time": self.observation_start_time,
            "observation_end_time": self.observation_end_time,
        }
        dataset.attrs["orbital_parameters"] = self._get_orbital_parameters()
        if self.include_raw_metadata:
            dataset.attrs["raw_metadata"] = reduce_mda(
                self.header, max_size=self.mda_max_array_size
            )

    def _get_orbital_parameters(self):
        orbital_parameters = {
            "projection_longitude": self.mda["projection_parameters"][
                "ssp_longitude"],
            "projection_latitude": 0.,
            "projection_altitude": self.mda["projection_parameters"]["h"],
            "satellite_nominal_longitude": self.header["15_DATA_HEADER"][
                "SatelliteStatus"]["SatelliteDefinition"][
                "NominalLongitude"],
            "satellite_nominal_latitude": 0.0
        }
        try:
            actual_lon, actual_lat, actual_alt = self.satpos
            orbital_parameters.update({
                "satellite_actual_longitude": actual_lon,
                "satellite_actual_latitude": actual_lat,
                "satellite_actual_altitude": actual_alt
            })
        except NoValidOrbitParams as err:
            logger.warning(err)
        return orbital_parameters

    @cached_property
    def satpos(self):
        """Get actual satellite position in geodetic coordinates (WGS-84).

        Evaluate orbit polynomials at the start time of the scan.

        Returns: Longitude [deg east], Latitude [deg north] and Altitude [m]
        """
        poly_finder = OrbitPolynomialFinder(self.header["15_DATA_HEADER"][
                                                "SatelliteStatus"]["Orbit"]["OrbitPolynomial"])
        orbit_polynomial = poly_finder.get_orbit_polynomial(self.start_time)
        return get_satpos(
            orbit_polynomial=orbit_polynomial,
            time=self.observation_start_time,
            semi_major_axis=self.mda["projection_parameters"]["a"],
            semi_minor_axis=self.mda["projection_parameters"]["b"]
        )


class ImageBoundaries:
    """Collect image boundary information."""

    def __init__(self, header, trailer, mda):
        """Initialize the class."""
        self._header = header
        self._trailer = trailer
        self._mda = mda

    def get_img_bounds(self, dataset_id, is_roi):
        """Get image line and column boundaries.

        Returns:
            Dictionary with the four keys 'south_bound', 'north_bound', 'east_bound' and 'west_bound',
            each containing a list of the respective line/column numbers of the image boundaries.

        Lists (rather than scalars) are returned since the HRV data in FES mode contain data from two windows/areas.
        """
        if dataset_id["name"] == "HRV" and not is_roi:
            img_bounds = self._get_hrv_actual_img_bounds()
        else:
            img_bounds = self._get_selected_img_bounds(dataset_id)

        self._check_for_valid_bounds(img_bounds)

        return img_bounds

    def _get_hrv_actual_img_bounds(self):
        """Get HRV (if not ROI) image boundaries from the ActualL15CoverageHRV information stored in the trailer."""
        hrv_bounds = self._trailer["15TRAILER"]["ImageProductionStats"]["ActualL15CoverageHRV"]

        img_bounds = {"south_bound": [], "north_bound": [], "east_bound": [], "west_bound": []}
        for hrv_window in ["Lower", "Upper"]:
            img_bounds["south_bound"].append(hrv_bounds["%sSouthLineActual" % hrv_window])
            img_bounds["north_bound"].append(hrv_bounds["%sNorthLineActual" % hrv_window])
            img_bounds["east_bound"].append(hrv_bounds["%sEastColumnActual" % hrv_window])
            img_bounds["west_bound"].append(hrv_bounds["%sWestColumnActual" % hrv_window])

            # Data from the upper hrv window are only available in FES mode
            if not self._mda["is_full_disk"]:
                break

        return img_bounds

    def _get_selected_img_bounds(self, dataset_id):
        """Get VISIR and HRV (if ROI) image boundaries from the SelectedRectangle information stored in the header."""
        sec15hd = self._header["15_SECONDARY_PRODUCT_HEADER"]
        south_bound = int(sec15hd["SouthLineSelectedRectangle"]["Value"])
        east_bound = int(sec15hd["EastColumnSelectedRectangle"]["Value"])

        if dataset_id["name"] == "HRV":
            nlines, ncolumns = self._get_hrv_img_shape()
            south_bound = self._convert_visir_bound_to_hrv(south_bound)
            east_bound = self._convert_visir_bound_to_hrv(east_bound)
        else:
            nlines, ncolumns = self._get_visir_img_shape()

        north_bound = south_bound + nlines - 1
        west_bound = east_bound + ncolumns - 1

        img_bounds = {"south_bound": [south_bound], "north_bound": [north_bound],
                      "east_bound": [east_bound], "west_bound": [west_bound]}

        return img_bounds

    def _get_hrv_img_shape(self):
        nlines = int(self._mda["hrv_number_of_lines"])
        ncolumns = int(self._mda["hrv_number_of_columns"])
        return nlines, ncolumns

    def _get_visir_img_shape(self):
        nlines = int(self._mda["number_of_lines"])
        ncolumns = int(self._mda["number_of_columns"])
        return nlines, ncolumns

    @staticmethod
    def _convert_visir_bound_to_hrv(bound):
        return 3 * bound - 2

    @staticmethod
    def _check_for_valid_bounds(img_bounds):
        len_img_bounds = [len(bound) for bound in img_bounds.values()]

        same_lengths = (len(set(len_img_bounds)) == 1)
        no_empty = (min(len_img_bounds) > 0)

        if not (same_lengths and no_empty):
            raise ValueError("Invalid image boundaries")


class Padder:
    """Padding of HRV, RSS and ROI data to full disk."""

    def __init__(self, dataset_id, img_bounds, is_full_disk):
        """Initialize the padder."""
        self._img_bounds = img_bounds
        self._is_full_disk = is_full_disk

        if dataset_id["name"] == "HRV":
            self._final_shape = (HRV_NUM_LINES, HRV_NUM_COLUMNS)
        else:
            self._final_shape = (VISIR_NUM_LINES, VISIR_NUM_COLUMNS)

    def pad_data(self, dataset):
        """Pad data to full disk with empty pixels."""
        logger.debug("Padding data to full disk")

        data_list = []
        for south_bound, north_bound, east_bound, west_bound in zip(*self._img_bounds.values()):
            nlines = north_bound - south_bound + 1
            data = self._extract_data_to_pad(dataset, south_bound, north_bound)
            padded_data = pad_data_horizontally(data, (nlines, self._final_shape[1]), east_bound, west_bound)
            data_list.append(padded_data)

        padded_data = da.vstack(data_list)

        # If we're dealing with RSS or ROI data, we also need to pad vertically in order to form a full disk array
        if not self._is_full_disk:
            padded_data = pad_data_vertically(padded_data, self._final_shape, south_bound, north_bound)

        return xr.DataArray(padded_data, dims=("y", "x"), attrs=dataset.attrs.copy())

    def _extract_data_to_pad(self, dataset, south_bound, north_bound):
        """Extract the data that shall be padded.

        In case of FES (HRV) data, 'dataset' contains data from two separate windows that
        are padded separately. Hence, we extract a subset of data.
        """
        if self._is_full_disk:
            data = dataset[south_bound - 1:north_bound, :].data
        else:
            data = dataset.data

        return data


def get_available_channels(header):
    """Get the available channels from the header information."""
    channels_str = header["15_SECONDARY_PRODUCT_HEADER"][
        "SelectedBandIDs"]["Value"]
    available_channels = {}

    for idx, char in zip(range(12), channels_str):
        available_channels[CHANNEL_NAMES[idx + 1]] = (char == "X")

    return available_channels


def has_archive_header(filename):
    """Check whether the file includes an ASCII archive header."""
    with generic_open(filename, mode="rb") as istream:
        return istream.read(36) == ASCII_STARTSWITH


def read_header(filename):
    """Read SEVIRI L1.5 native header."""
    dtype = get_native_header(has_archive_header(filename))
    hdr = fromfile(filename, dtype=dtype, count=1)
    return recarray2dict(hdr)


def _get_array(filename=None, hdr_size=None, block_info=None):
    """Get the numpy array for the SEVIRI data."""
    output_block_info = block_info[None]
    data_dtype = output_block_info["dtype"]
    return fromfile(
        filename,
        dtype=data_dtype,
        offset=hdr_size + output_block_info["array-location"][0][0] * data_dtype.itemsize,
        count=output_block_info["chunk-shape"][0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Demo FCI data download."""

import pathlib
import tarfile
import tempfile

from satpy import config

from . import utils

_fci_uncompressed_nominal = (
    "https://sftp.eumetsat.int/public/folder/UsCVknVOOkSyCdgpMimJNQ/"
    "User-Materials/Test-Data/MTG/MTG_FCI_L1C_Enhanced-NonN_TD-272_May2020/"
    "FCI_1C_UNCOMPRESSED_NOMINAL.tar.gz")


def download_fci_test_data(base_dir=None):
    """Download FCI test data.

    Download the nominal FCI test data from July 2020.
    """
    subdir = get_fci_test_data_dir(base_dir=base_dir)
    with tempfile.TemporaryDirectory() as td:
        nm = pathlib.Path(td) / "fci-test-data.tar.gz"
        utils.download_url(_fci_uncompressed_nominal, nm)
        return _unpack_tarfile_to(nm, subdir)


def get_fci_test_data_dir(base_dir=None):
    """Get directory for FCI test data."""
    base_dir = base_dir or config.get("demo_data_dir", ".")
    return pathlib.Path(base_dir) / "fci" / "test_data"


def _unpack_tarfile_to(filename, subdir):
    """Unpack content of tarfile in filename to subdir."""
    with tarfile.open(filename, mode="r:gz") as tf:
        contents = tf.getnames()
        tf.extractall(path=subdir, filter="data")
    return contents



# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Utility to generate a CF-compliant Datasets."""
import logging
import warnings

import xarray as xr

from satpy.writers.cf_writer import CF_DTYPES, CF_VERSION

logger = logging.getLogger(__name__)


def _get_extra_ds(dataarray, keys=None):
    """Get the ancillary_variables DataArrays associated to a dataset."""
    dict_datarrays = {}
    # Retrieve ancillary variable datarrays
    for ancillary_dataarray in dataarray.attrs.get("ancillary_variables", []):
        ancillary_variable = ancillary_dataarray.name
        if keys and ancillary_variable not in keys:
            keys.append(ancillary_variable)
            dict_datarrays.update(_get_extra_ds(ancillary_dataarray, keys=keys))
    # Add input dataarray
    dict_datarrays[dataarray.attrs["name"]] = dataarray
    return dict_datarrays


def _get_group_dataarrays(group_members, list_dataarrays):
    """Yield DataArrays that are part of a specific group."""
    return [da for da in list_dataarrays if da.attrs["name"] in group_members]


def _get_groups(groups, list_datarrays):
    """Return a dictionary with the list of xr.DataArray associated to each group.

    If no groups (groups=None), return all DataArray attached to a single None key.
    Else, collect the DataArrays associated to each group.
    """
    if groups is None:
        return {None: list_datarrays}

    return {group_name: _get_group_dataarrays(group_members, list_datarrays)
            for group_name, group_members in groups.items()}


def _collect_cf_dataset(list_dataarrays,
                        epoch=None,
                        flatten_attrs=False,
                        exclude_attrs=None,
                        include_lonlats=True,
                        pretty=False,
                        include_orig_name=True,
                        numeric_name_prefix="CHANNEL_"):
    """Process a list of xr.DataArray and return a dictionary with CF-compliant xr.Dataset.

    Args:
        list_dataarrays (list): List of DataArrays to make CF compliant and merge into an xr.Dataset.
        epoch (str, Optional): Reference time for encoding the time coordinates.
            Example format: "seconds since 1970-01-01 00:00:00".
            If None, the default reference time is defined using `from satpy.cf.coords import EPOCH`.
        flatten_attrs (bool, Optional): If True, flatten dict-type attributes.
        exclude_attrs (list, Optional): List of xr.DataArray attribute names to be excluded.
        include_lonlats (bool, Optional): If True, includes 'latitude' and 'longitude' coordinates also for a
            satpy.Scene defined on an AreaDefinition.
            If the 'area' attribute is a SwathDefinition, it always includes latitude and longitude coordinates.
        pretty (bool, Optional): Don't modify coordinate names, if possible.
            Makes the file prettier, but possibly less consistent.
        include_orig_name (bool, Optional): Include the original dataset name as a variable attribute in the xr.Dataset.
        numeric_name_prefix (str, Optional): Prefix to add to each variable with a name starting with a digit.
            Use '' or None to leave this out.

    Returns:
        xarray.Dataset: A partially CF-compliant xr.Dataset.
    """
    from satpy.cf.area import area2cf
    from satpy.cf.coords import (
        add_coordinates_attrs_coords,
        check_unique_projection_coords,
        ensure_unique_nondimensional_coords,
        has_projection_coords,
    )
    from satpy.cf.data_array import make_cf_data_array

    # Create dictionary of input datarrays
    # --> Since keys=None, it doesn't never retrieve ancillary variables !!!
    dict_dataarrays = {}
    for dataarray in list_dataarrays:
        dict_dataarrays.update(_get_extra_ds(dataarray))

    # Check if one DataArray in the collection has 'longitude' or 'latitude'
    got_lonlats = has_projection_coords(dict_dataarrays)

    # Sort dictionary by keys name
    dict_dataarrays = dict(sorted(dict_dataarrays.items()))

    dict_cf_dataarrays = {}
    for dataarray in dict_dataarrays.values():
        dataarray_type = dataarray.dtype
        if dataarray_type not in CF_DTYPES:
            warnings.warn(
                f"dtype {dataarray_type} not compatible with {CF_VERSION}.",
                stacklevel=3
            )
        # Deep copy the datarray since adding/modifying attributes and coordinates
        dataarray = dataarray.copy(deep=True)

        # Add CF-compliant area information from the pyresample area
        # - If include_lonlats=True, add latitude and longitude coordinates
        # - Add grid_mapping attribute to the DataArray
        # - Return the CRS DataArray as first list element
        # - Return the CF-compliant input DataArray as second list element
        try:
            list_new_dataarrays = area2cf(dataarray,
                                          include_lonlats=include_lonlats,
                                          got_lonlats=got_lonlats)
        except KeyError:
            list_new_dataarrays = [dataarray]

        # Ensure each DataArray is CF-compliant
        # --> NOTE: Here the CRS DataArray is repeatedly overwrited
        # --> NOTE: If the input list_dataarrays have different pyresample areas with the same name
        #           area information can be lost here !!!
        for new_dataarray in list_new_dataarrays:
            new_dataarray = make_cf_data_array(new_dataarray,
                                               epoch=epoch,
                                               flatten_attrs=flatten_attrs,
                                               exclude_attrs=exclude_attrs,
                                               include_orig_name=include_orig_name,
                                               numeric_name_prefix=numeric_name_prefix)
            dict_cf_dataarrays[new_dataarray.name] = new_dataarray

    # Check all DataArrays have same projection coordinates
    check_unique_projection_coords(dict_cf_dataarrays)

    # Add to DataArrays the coordinates specified in the 'coordinates' attribute
    # - Deal with the 'coordinates' attributes indicating lat/lon coords
    # - The 'coordinates' attribute is dropped from each DataArray
    dict_cf_dataarrays = add_coordinates_attrs_coords(dict_cf_dataarrays)

    # Ensure non-dimensional coordinates to be unique across DataArrays
    # --> If not unique, prepend the DataArray name to the coordinate
    # --> If unique, does not prepend the DataArray name only if pretty=True
    # --> 'longitude' and 'latitude' coordinates are not prepended
    dict_cf_dataarrays = ensure_unique_nondimensional_coords(dict_cf_dataarrays, pretty=pretty)

    # Create a xr.Dataset
    ds = xr.Dataset(dict_cf_dataarrays)
    return ds


def collect_cf_datasets(list_dataarrays,
                        header_attrs=None,
                        exclude_attrs=None,
                        flatten_attrs=False,
                        pretty=True,
                        include_lonlats=True,
                        epoch=None,
                        include_orig_name=True,
                        numeric_name_prefix="CHANNEL_",
                        groups=None):
    """Process a list of xr.DataArray and return a dictionary with CF-compliant xr.Datasets.

    If the xr.DataArrays does not share the same dimensions, it creates a collection
    of xr.Datasets sharing the same dimensions.

    Args:
        list_dataarrays (list): List of DataArrays to make CF compliant and merge into groups of xr.Datasets.
        header_attrs (dict): Global attributes of the output xr.Dataset.
        epoch (str, Optional): Reference time for encoding the time coordinates.
            Example format: "seconds since 1970-01-01 00:00:00".
            If None, the default reference time is retrieved using `from satpy.cf.coords import EPOCH`.
        flatten_attrs (bool, Optional): If True, flatten dict-type attributes.
        exclude_attrs (list, Optional): List of xr.DataArray attribute names to be excluded.
        include_lonlats (bool, Optional): If True, includes 'latitude' and 'longitude' coordinates also
            for a satpy.Scene defined on an AreaDefinition.
            If the 'area' attribute is a SwathDefinition, it always includes latitude and longitude coordinates.
        pretty (bool, Optional): Don't modify coordinate names, if possible.
            Makes the file prettier, but possibly less consistent.
        include_orig_name (bool, Optional): Include the original dataset name as a variable attribute in the xr.Dataset.
        numeric_name_prefix (str, Optional): Prefix to add to each variable with a name starting with a digit.
            Use '' or None to leave this out.
        groups (dict, Optional): Group datasets according to the given assignment:
            `{'<group_name>': ['dataset_name1', 'dataset_name2', ...]}`.
            Used to create grouped netCDFs using the CF_Writer. If None, no groups will be created.

    Returns:
        tuple: A tuple containing:
            - grouped_datasets (dict): A dictionary of CF-compliant xr.Dataset: {group_name: xr.Dataset}.
            - header_attrs (dict): Global attributes to be attached to the xr.Dataset / netCDF4.
    """
    from satpy.cf.attrs import preprocess_header_attrs
    from satpy.cf.coords import add_time_bounds_dimension

    if not list_dataarrays:
        raise RuntimeError("None of the requested datasets have been "
                           "generated or could not be loaded. Requested "
                           "composite inputs may need to have matching "
                           "dimensions (eg. through resampling).")

    header_attrs = preprocess_header_attrs(header_attrs=header_attrs,
                                           flatten_attrs=flatten_attrs)

    # Retrieve groups
    # - If groups is None: {None: list_dataarrays}
    # - if groups not None: {group_name: [xr.DataArray, xr.DataArray ,..], ...}
    # Note: if all dataset names are wrong, behave like groups = None !
    grouped_dataarrays = _get_groups(groups, list_dataarrays)
    is_grouped = len(grouped_dataarrays) >= 2

    # If not grouped, add CF conventions.
    # - If 'Conventions' key already present, do not overwrite !
    if "Conventions" not in header_attrs and not is_grouped:
        header_attrs["Conventions"] = CF_VERSION

    # Create dictionary of group xr.Datasets
    # --> If no groups (groups=None) --> group_name=None
    grouped_datasets = {}
    for group_name, group_dataarrays in grouped_dataarrays.items():
        ds = _collect_cf_dataset(
            list_dataarrays=group_dataarrays,
            epoch=epoch,
            flatten_attrs=flatten_attrs,
            exclude_attrs=exclude_attrs,
            include_lonlats=include_lonlats,
            pretty=pretty,
            include_orig_name=include_orig_name,
            numeric_name_prefix=numeric_name_prefix)

        if not is_grouped:
            ds.attrs = header_attrs

        if "time" in ds:
            ds = add_time_bounds_dimension(ds, time="time")

        grouped_datasets[group_name] = ds
    return grouped_datasets, header_attrs
